# -*- coding: utf-8 -*-
"""
Utility functions for randomization statistics

Function list
-------------
- resamples_to_pvalue : Compute p value from observed and resampled statistic values
- confint_to_indexes : Returns indexes into resampled stats corresponding to given confints
- jackknife_to_pseudoval : Compute single-trial pseudovalues from leave-one-out jackknife estimates

Function reference
------------------
"""
import numpy as np

from spynal.randstats.helpers import _tail_to_compare


def resamples_to_pvalue(stat_obs, stat_resmp, axis=0, tail='both'):
    """
    Compute p value with given tail from observed and resampled values of a statistic

    Parameters
    ----------
    stat_obs : ndarray, shape=(...,1,...)
        Statistic values for actual observed data

    stat_resmp : ndarray, shape=(...,n_resamples,...)
        Statistic values for randomly resampled data

    axis : int, default: 0
        Axis in `stat_resmp` corresponding to distinct resamples
        (should correspond to a length=1 axis in `stat_obs`)

    tail : {'both','right','left'}, default: 'both' (2-tailed test)
        Specifies tail of test to perform:

        - 'both'  : 2-tail test -- test for abs(stat_obs) > abs(stat_resmp)
        - 'right' : right-sided 1-tail test -- tests for stat_obs > stat_resmp
        - 'left'  : left-sided 1-tail test -- tests for stat_obs < stat_resmp

    Returns
    -------
    p : ndarray, shape=(...,1,...)
        p values from resampling test. Same size as `stat_obs`.
    """
    compare_func = _tail_to_compare(tail)           

    n_resamples = stat_resmp.shape[axis]

    # Count number of resampled stat values more extreme than observed value
    p = np.sum(compare_func(stat_obs,stat_resmp), axis=axis, keepdims=True)

    # p value is proportion of samples failing criterion (+1 for observed stat)
    return (p + 1) / (n_resamples + 1)


def confint_to_indexes(confint, n_resamples):
    """
    Return indexes into set of resamples corresponding to given confidence interval

    Typically used for bootstrap resampled confidence intervals. Could be used for jackknifes.

    Parameters
    ----------
    confint : float
        Desired confidence interval, in range 0-1. eg, for 99% confidence interval, input 0.99

    n_resamples : int
        Number of resamples (eg bootstraps)

    Returns
    -------
    conf_indexes : list[int], shape=(2,)
        Indexes into sorted resamples corresponding to [lower,upper] confidence interval
    """
    max_interval = 1 - 2.0/n_resamples
    assert (confint <= max_interval) or np.isclose(confint,max_interval), \
        ValueError("Requested confint too large for given number of resamples (max = %.3f)" \
                    % max_interval)

    return [round(n_resamples * (1-confint)/2) - 1,
            round(n_resamples - (n_resamples * (1-confint)/2)) - 1]


def jackknife_to_pseudoval(x, xjack, n):
    """
    Compute single-trial pseudovalues from leave-one-out jackknife estimates

    Parameters
    ----------
    x : ndarray, shape=Any
        Statistic computed on full observed data. Any arbitrary shape.

    xjack : ndarray, shape=Any
        Statistic computed on jackknife resampled data

    n : int
        Number of observations (trials) used to compute x

    Returns
    -------
    pseudo : ndarray, shape=Any
        Single-trial jackknifed pseudovalues. Same shape as xjack.
    """
    return n*x - (n-1)*xjack



""" Unit tests for matIO.py module """
import pytest
import os
import numpy as np
import pandas as pd

import tempfile

from spynal.tests.data_fixtures import MISSING_ARG_ERRS
from spynal.matIO import loadmat, load, whomat, who, savemat, save


# =============================================================================
# Unit tests
# =============================================================================
@pytest.mark.parametrize('version', [('v7'), ('v73')])
def test_loadmat(version):
    """ Unit tests for loadmat function in matIO module """
    filename = _set_filename(version)
    
    variables = ['integer','floating','boolean','string', 'num_array','cell_array',
                 'gen_struct','table_struct']
    # scipy.io returns v7 Matlab logicals as ints (not bools) and not much we can do about it
    bool_type = int if version == 'v7' else bool
    
    extra_args = dict(asdict=True, extract_items=True, verbose=False)

    # Test data is loaded correctly when all variables loaded into a dict
    data = loadmat(filename, **extra_args)
    # _print_data_summary(data)
    _variable_tests(data, table_type=dict, bool_type=bool_type, extract_item=True)

    # Test function alias
    data = load(filename, **extra_args)
    _variable_tests(data, table_type=dict, extract_item=True, bool_type=bool_type)

    # Test extract_item=False
    extra_args['extract_items'] = False
    data = load(filename, **extra_args)
    _variable_tests(data, table_type=dict, extract_item=False, bool_type=bool_type)
    extra_args['extract_items'] = True
    
    # Test data is loaded correctly when all variables loaded into separate variables
    extra_args['asdict'] = False    
    integer, floating, boolean, string, num_array, cell_array, gen_struct, table_struct = \
        loadmat(filename, variables=variables, **extra_args)
    data = {'integer':integer, 'floating':floating, 'boolean':boolean, 'string':string,
            'num_array':num_array, 'cell_array':cell_array,
            'gen_struct':gen_struct, 'table_struct':table_struct}
    _variable_tests(data, dict, bool_type=bool_type, extract_item=True)
    extra_args['asdict'] = True    

    # Test for correct loading when variables to load are specified (as list or single name)
    data = loadmat(filename, variables=['integer','string'], **extra_args)
    assert list(data.keys()) == ['integer','string']
    assert isinstance(data['integer'], int)
    assert data['integer'] == 1
    assert isinstance(data['string'], str)
    assert data['string'] == 'abc'

    data = loadmat(filename, variables='integer', **extra_args)
    assert list(data.keys()) == ['integer']
    assert isinstance(data['integer'], int)
    assert data['integer'] == 1

    # Test for correct loading of table-like struct into Pandas DataFrame
    data = loadmat(filename, typemap={'table_struct':'DataFrame'}, **extra_args)
    _variable_tests(data, pd.DataFrame, bool_type)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        data = loadmat(filename, **extra_args, foo=None)
        

@pytest.mark.parametrize('version', [('v7'), ('v73')])
def test_whomat(version):
    """ Unit tests for whomat function in matIO module """
    filename = _set_filename(version)
    
    variables = ['boolean', 'cell_array', 'floating', 'gen_struct', 'integer',
                 'num_array', 'string', 'table_struct']
    
    # Basic test of functionality
    test_vbls = sorted(whomat(filename, verbose=False))
    assert test_vbls == variables
    
    # Test function alias
    test_vbls = sorted(who(filename, verbose=False))
    assert test_vbls == variables

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        data = whomat(filename, verbose=False, foo=None)


@pytest.mark.parametrize('version', [('v7'), ('v73')])
def test_savemat(version):
    _version = 7 if version == 'v7' else 7.3
    
    # Load variables from test file
    filename = _set_filename(version)    
    variables = loadmat(filename, asdict=True, verbose=False)

    # Save variables out to temporary file        
    with tempfile.TemporaryDirectory() as temp_folder:
        # Basic test of function
        filename = os.path.join(temp_folder, 'test_matfile_'+version+'.mat')
        savemat(filename, variables, version=_version)

        # Test function alias
        save(filename, variables, version=_version)

        # Ensure that passing a nonexistent/misspelled kwarg raises an error
        with pytest.raises(MISSING_ARG_ERRS):
            savemat(filename, variables, version=_version, foo=None)
            
                
def test_imports():
    """ Test different import methods for matIO module """
    # Import entire package
    import spynal
    spynal.matIO.matIO.load    
    spynal.matIO.load
    # Import module
    import spynal.matIO as matIO
    matIO.matIO.load
    matIO.load
    # Import specific function from module
    from spynal.matIO import load
    load
    # Import specific function from module
    from spynal.matIO.matIO import load
    load
    
    
# =============================================================================
# Helper functions for unit tests
# ============================================================================= 
def _set_filename(version):
    """ Set filename for test file independent of current dir """
    # HACK Make this work when run from top-level Python dir in VS Code or from tests dir in terminal
    cwd = os.getcwd()
    if cwd.endswith('tests'):   load_dir = r'./'
    else:                       load_dir = r'./spynal/spynal/tests'
    return os.path.join(load_dir, 'testing_datafile_' + version + '.mat')

           
def _variable_tests(data, table_type=dict, bool_type=bool, extract_item=True):
    """ Set of tests to run on loadmat-loaded variables """
    assert isinstance(data, dict)

    # Test scalar and array variables
    _scalar_tests(data, bool_type=bool_type, extract_item=extract_item)
    _array_tests(data)

    # Test generic struct variable -- loaded as dict
    _scalar_tests(data['gen_struct'], bool_type=bool_type, extract_item=extract_item)
    _array_tests(data['gen_struct'])

    # Test table-like struct variable -- loaded as dict or Pandas DataFrame
    assert isinstance(data['table_struct'], table_type)
    col_type = np.ndarray if isinstance(data['table_struct'],dict) else pd.Series

    assert isinstance(data['table_struct']['num_array'], col_type)
    assert np.issubdtype(data['table_struct']['num_array'].dtype, float)
    assert data['table_struct']['num_array'].shape == (4,)
    assert (data['table_struct']['num_array'] == np.arange(1,5)+0.1).all()

    assert isinstance(data['table_struct']['cell_array'], col_type)
    assert np.issubdtype(data['table_struct']['cell_array'].dtype, object)
    assert data['table_struct']['cell_array'].shape == (4,)
    assert (data['table_struct']['cell_array'] == np.asarray(['abc','def','gh','ij'])).all()


def _scalar_tests(data, bool_type=bool, extract_item=True):
    """ Tests for scalar variables """
    if extract_item:
        assert isinstance(data['integer'], int)
        assert data['integer'] == 1

        assert isinstance(data['floating'], float)
        assert data['floating'] == 1.1

        assert isinstance(data['boolean'], bool_type)
        assert data['boolean'] == True

        assert isinstance(data['string'], str)
        assert data['string'] == 'abc'
    
    else:
        assert isinstance(data['integer'], np.ndarray) and np.issubdtype(data['integer'].dtype, np.integer)
        assert np.array_equal(data['integer'], np.array(1, dtype=np.int16))

        assert isinstance(data['floating'], np.ndarray) and np.issubdtype(data['floating'].dtype, float)
        assert np.array_equal(data['floating'], np.array(1.1, dtype=float))

        if bool_type == int: bool_type = np.integer 
        assert isinstance(data['boolean'], np.ndarray) and np.issubdtype(data['boolean'].dtype, bool_type)
        assert np.array_equal(data['boolean'], np.array(True, dtype=bool))
        
        assert isinstance(data['string'], np.ndarray) and np.issubdtype(data['string'].dtype, object)
        assert np.array_equal(data['string'], np.array('abc', dtype=object))
        

def _array_tests(data):
    """ Tests for array variables """
    assert isinstance(data['num_array'], np.ndarray)
    assert np.issubdtype(data['num_array'].dtype, float)
    assert data['num_array'].shape == (4,3,2)
    assert (data['num_array'] == np.arange(1,25).reshape((4,3,2),order='F')+0.1).all()

    assert isinstance(data['cell_array'], np.ndarray)
    assert np.issubdtype(data['cell_array'].dtype, object)
    assert data['cell_array'].shape == (4,6)
    assert (data['cell_array'] == np.tile(np.asarray(['abc','def','gh','ij'])[:,np.newaxis], (1,6))).all()


def _print_data_summary(data):
    """ Print summary of key/value pairs in data struct """
    print(data.keys())
    for key in data:
        if np.isscalar(data[key]):              size = 1
        elif isinstance(data[key],np.ndarray):  size = data[key].shape
        elif hasattr(data[key], '__len__'):     size = len(data[key])
        else:                                   size = '???'
        print(key, type(data[key]), size)
        print(data[key])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Functions for generating common data plots and plotting-related utilities

Overview
--------
High-level interfaces to create plot types commonly used in neuroscience, including heat maps,
line/curve plots with tranparent fills indicating errors, and offset lineseries plots (eg for
plotting series of ephys traces or evoked potentials).

Also contains some utility functions useful in generating plots.

Built using `matplotlib.pyplot` functions.


Function list
-------------
Plot-generating functions
^^^^^^^^^^^^^^^^^^^^^^^^^
- plot_line_with_error_fill :   Plot 1d data as line(s) w/ transparent fill(s) to indicate errors
- plot_heatmap :                Plot 2d data as a heatmap (pseudocolor) plot
- plot_lineseries :             Plot 2d data as series of offset lines

Plotting utilities
^^^^^^^^^^^^^^^^^^
- full_figure :                 Create full-screen figure
- savefig :                     Save figure to file in ~WYSIWYG manner
- make_colormap :               Create custom colormap and register name for further use
- colorbar :                    Create colorbar without messing up parent axis size/shape
- plot_markers :                Plot set of markers (eg to mark trial event times) on given axis(s)

Function reference
------------------
"""
# Created on Tue Nov 23 14:22:47 2021
#
# @author: sbrincat
import os
from warnings import warn
import numpy as np
import matplotlib.pyplot as plt

from matplotlib import get_backend
from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from matplotlib.image import AxesImage
from matplotlib.patches import Polygon
from matplotlib.colors import ListedColormap, LinearSegmentedColormap

from spynal.utils import isnumeric, isarraylike
from spynal.helpers import _merge_dicts

# Lambda returns list of all settable attributes of given plotting object
# Find all methods starting with 'set_***', strip out the 'set_', and place in a list
_settable_attributes = lambda obj: ['_'.join(attr.split('_')[1:]) for attr in dir(obj) \
                                    if attr.startswith('set_')]

# Create list of all settable attributes/parameters of plotting objects/functions used in module
AXES_PARAMS = _settable_attributes(Axes)
PLOT_PARAMS = ['scalex', 'scaley'] + _settable_attributes(Line2D)
FILL_PARAMS = _settable_attributes(Polygon)
# For some reason, imshow has several settable params that aren't AxesImage attributes...
# this seems to be a way to get them (though not for other functions...?)
IMSHOW_PARAMS = ['aspect','origin'] + _settable_attributes(AxesImage)


# =============================================================================
# Functions to generate specific plot types
# =============================================================================
def plot_line_with_error_fill(x, data, err=None, ax=None, color=None, events=None, **kwargs):
    """
    Plot 1d data as line plot(s) +/- error(s) as semi-transparent fill(s) in given axis

    Can plot multiple lines/errors with same x-axis values in same axis by inputting
    multiple data/error series

    Uses :func:`plt.plot` and :func:`plt.fill`

    Parameters
    ----------
    x : array-like, shape=(n,)
        x-axis sampling vector for both `data` and `err`.

    data : array-like, shape=(n,) or (n_lines,n)
        Values to plot on y-axis as line(s) (typically means or medians).
        To plot multiples lines with the same x-axis values, input a 2D array where
        each row will be plotted as a separate line.

    err : array-like, shape=(n,) or (n_lines,n) or (2*n_lines,n), default: (no errorbars)
        Error values (SEMs/confints/etc.) to plot as semi-transparent fill around line(s).

        If vector-valued or (n_lines,n) array, errors are treated as 1-sided (like SEMs),
        and `data[j,:]` +/- `err[j,:]` is plotted for each line j.

        If given as (2*n_lines,n) array, it's treated as 2-sided [upper; lower] error ranges
        (like confidence intervals), with the odd rows = upper and even rows = lower errors
        corresponding to each line.

        If err=None [default], only the data line(s) are plotted without error fills
        (to simplify calling code with optional errors).

    color : Color spec or (n_lines,) of Color specs, default: (standard matplotlib color order)
        Color to plot each line(s) and error fill(s) in.
        Can input either a single color spec to use for all lines/fills, or 1 per line/fill pair.
        Can input in any of the ways matplotlib colors are defined (strings, 3-tuples, etc.)

    events : callable or array-like, shape=(n_events,)
        List of event values (eg times) to plot as markers on x-axis
        -or- callable function that will just plot the event markers itself.
        See :func:`plot_markers` for details.

    **kwargs
        Any additional keyword args are interpreted as parameters of :func:`plt.axes`
        (settable Axes object attributes), :func:`plt.plot` (Line2D object attributes),
        or :func:`plt.fill` (Polygon object attributes) and passsed to the proper function.
        A few commonly used options, with custom defaults:

        linewidth : scalar, default: 1.5
            Width of all plotted data line(s)

        alpha : float, range=[0,1], default: 0.25
            Transparency alpha value for plotting all error fill(s).
            1=fully opaque, 0=fully transparent.

    Returns
    -------
    lines : list of Line2D objects
        ax.plot output. Allows access to line properties of line.

    patches : list of Polygon objects
        ax.fill output. Allows access to patch properties of fill.

    ax : Axis object
        Axis plotted into
    """
    x = np.asarray(x)
    data = np.asarray(data)
    if data.ndim == 1: data = data[np.newaxis,:] # Convert 1d (n,) data -> (1,n) to simplify code
    n_lines = data.shape[0]

    assert data.ndim == 2, \
        ValueError("data must be 1-d (or 2-d for multiple lines) (%d-d data given)" % data.ndim)
    assert data.shape[1] == len(x), \
        ValueError("data (%d) and x (%d) should have same length" % (data.shape[1],len(x)))

    if err is not None:
        err = np.asarray(err)
        if err.ndim == 1: err = err[np.newaxis,:]

        assert err.shape[1] == len(x), \
            ValueError("err.shape[1] (%d) and x (%d) should have same length" % (err.shape[1],len(x)))
        assert err.shape[0] in [n_lines,2*n_lines], \
            ValueError("err must be input as (n,) vector or (2*n_lines,n) array of upper;lower errors")

        # Convert errors to 2-sided upper and lower error bounds to simplify code
        if err.shape[0] == n_lines: upper, lower = data+err, data-err
        else:                       upper, lower = err[0::2,:], err[1::2,:]

        ylim = (lower.min(), upper.max())

    else:
        ylim = (data.min(), data.max())

    # Default ylim to data range +/- 5%
    ylim = (ylim[0]-0.05*np.diff(ylim), ylim[1]+0.05*np.diff(ylim))
    xlim = (x.min(),x.max())

    # Set axis to plot into (default to current axis)
    if ax is None: ax = plt.gca()

    # Sort any keyword args to their appropriate plotting object
    axes_args, plot_args, fill_args = _hash_kwargs(kwargs, [AXES_PARAMS, PLOT_PARAMS, FILL_PARAMS])
    # Merge any input parameters with default values
    axes_args = _merge_dicts(dict(xlim=xlim, ylim=ylim), axes_args)
    plot_args = _merge_dicts(dict(linewidth=1.5), plot_args)
    fill_args = _merge_dicts(dict(alpha=0.25), fill_args)

    # Set plotting colors, including defaults
    color = _set_plot_colors(color, n_lines)

    ax.set(**axes_args) # Set axes parameters

    # Plot event markers (if input)
    if events is not None:
        if callable(events):    events()
        else:                   plot_markers(events, axis='x', ax=ax,
                                             xlim=axes_args['xlim'], ylim=axes_args['ylim'])

    # Plot line(s) and error fill(s) if input
    lines = []
    patches = []
    for j in range(n_lines):
        if err is not None:
            patch = ax.fill(np.hstack((x,np.flip(x))),
                            np.hstack((upper[j,:], np.flip(lower[j,:]))),
                            facecolor=color[j], **fill_args)
            patches.append(patch)

        line = ax.plot(x, data[j,:], '-', color=color[j], **plot_args)
        lines.append(line)

    return lines, patches, ax


def plot_heatmap(x, y, data, ax=None, clim=None, events=None, **kwargs):
    """
    Plot 2D data as a heatmap (aka pseudocolor) plot in given axis

    Uses :func:`plt.imshow`

    Parameters
    ----------
    x : array-like, shape=(n_x,)
        Sampling vector for data dimension to be plotted along x-axis

    y : array-like, shape=(n_y,)
        Sampling vector for data dimension to be plotted along y-axis

    data  : ndarray, shape=(n_y,n_x)
        Data to plot on color axis. NOTE: Data array must be 2d,
        with data to be plotted on y-axis the first dimension and the x-axis data 2nd.

    ax : Pyplot Axis object, default: plt.gca() (current axis)
        Axis to plot into.

    clim : array-like, shape=(2,), default: (data.min(),data.max()) (full range of data)
        [low,high] limits of color axis

    events : callable or array-like, shape=(n_events,)
        List of event values (eg times) to plot as markers on x-axis
        -or- callable function that will just plot the event markers itself.
        See :func:`plot_markers` for details.

    **kwargs
        Any additional keyword args are interpreted as parameters of :func:`plt.axes`
        (settable Axes object attributes) or :func:`plt.imshow` (AxesImage object attributes).
        A few commonly used options, with custom defaults:

        cmap  : str | Colormap object. default: 'viridis' (linear dark-blue to yellow colormap)
            Colormap to plot heatmap in, given either as name of matplotlib colormap or custom
            matplotlib.colors.Colormap object instance.

        origin : {'lower','upper'}, default: 'lower'
            Where 1st value in data is plotted along y-axis: 'lower'=bottom, 'upper'=top.

    Returns
    -------
    img : AxesImage object
        Output of ax.imshow(). Allows access to image properties.

    ax : Axis object
        Axis plotted into.
    """
    x = np.asarray(x)
    y = np.asarray(y)

    assert data.ndim == 2, ValueError("data must be 2-dimensional (%d-d data given)" % data.ndim)
    assert data.shape == (len(y),len(x)), \
        ValueError("data (%d,%d) must have dimensions (len(y),len(x)) = (%d,%d)" \
                    % (*data.shape,len(y),len(x)))

    # Set axis to plot into (default to current axis)
    if ax is None: ax = plt.gca()
    # Default color range to data min/max
    if clim is None: clim = (data.min(), data.max())

    # Find sampling intervals for x, y axes
    dx = np.diff(x).mean() if len(x) > 1 else 1
    dy = np.diff(y).mean() if len(y) > 1 else 1
    # Set default plotting extent for each axis = full sampling range +/- 1/2 sampling interval
    # This allows for viewing the entire cells at the edges of the plot, which sometimes makes
    # a difference for sparsely sampled dimensions
    xlim = [x[0]-dx/2, x[-1]+dx/2]
    ylim = [y[0]-dy/2, y[-1]+dy/2]

    # Sort any keyword args to their appropriate plotting object
    axes_args, imshow_args = _hash_kwargs(kwargs, [AXES_PARAMS, IMSHOW_PARAMS])
    # Merge any input parameters with default values
    axes_args = _merge_dicts(dict(xlim=xlim, ylim=ylim), axes_args)
    imshow_args = _merge_dicts(dict(extent=[*xlim,*ylim], vmin=clim[0], vmax=clim[1],
                                    cmap='viridis', origin='lower', aspect='auto',
                                    interpolation='none'), imshow_args)

    img = ax.imshow(data, **imshow_args)

    ax.set(**axes_args) # Set axes parameters

    # TODO Fix this
    # Have to manually invert y-axis tick labels if plotting w/ origin='upper'
    # if origin == 'upper':
    #     yticks = ax.get_yticks()
    #     idxs = (yticks[0] >= ylim[0]) & (yticks[0] <= ylim[1])
    #     yticks = yticks[0][idxs], np.asarray(yticks[1][idxs])
    #     plt.set_yticks(yticks[0])
    #     plt.set_yticklabels(np.flip(yticks[1]))

    # Plot event markers (if input)
    if events is not None:
        if callable(events):    events()
        else:                   plot_markers(events, axis='x', ax=ax,
                                             xlim=axes_args['xlim'], ylim=axes_args['ylim'])

    return img, ax


def plot_lineseries(x, y, data, ax=None, scale=1.5, color='C0', origin='upper',
                    events=None, **kwargs):
    """
    Plot 2d data as series of vertically-offset line plots with same x-axis values

    Used for example when plotting time-series traces from multiple electrodes on a linear probe.

    Uses :func:`plt.plot`

    Parameters
    ----------
    x : array-like, shape=(n_x,)
        Sampling vector for data dimension to be plotted on x-axis. This will often be timepoints.

    y : array-like, shape=(n_y,)
        Sampling vector for data dimension to be plotted along y-axis. Can be either numeric or
        string labels ot plot for y-axis ticklabels (one at each plotted line).
        This will often be channel numbers or channel labels.

    data : ndarray, shape=(n_y,n_x)
        Data to plot as vertically-offset line series. NOTE: Data array must be 2d,
        with data to be plotted on y-axis the first dimension and the x-axis data 2nd.

    ax :  Pyplot Axis object, default: plt.gca() (current axis)
         Axis to plot into

    scale : float, default: 1.5
        Scale factor for plotting data. 1 = max range of entire data maps to offset between
        successive lines in plot; >1 = extends farther; <1 = scaled smaller than offset.

    color : Color spec | (n_y,) of Color specs, default: 'C0' (blue for all lines)
        Color(s) to use to plot each line in line series.
        Can input either a single color spec to use for all lines/fills, or 1 per line/fill pair.
        Can input in any of the ways matplotlib colors are defined (strings, 3-tuples, etc.)

    origin : {'lower','upper'}, default: 'upper'
        Where 1st value in data is plotted along y-axis;'lower'=bottom, 'upper'=top.
        Default order plots in same order as a probe numbered from topmost contact.

    events : callable or array-like, shape=(n_events,)
        List of event values (eg times) to plot as markers on x-axis
        -or- callable function that will just plot the event markers itself.
        See :func:`plot_markers` for details.

    **kwargs
        Any additional keyword args are interpreted as parameters of :func:`plt.axes`
        (settable Axes object attributes) or :func:`plt.plot` (Line2D object attributes)
        and passsed to the proper function.
        A few commonly used options, with custom defaults:

        linewidth : scalar, default: 1
            Width of all plotted data line

    Returns
    -------
    lines : list of Line2D objects
        ax.plot output. Allows access to line properties of line.

    ax : Axis object
        Axis plotted into.
    """
    x = np.asarray(x)
    y = np.asarray(y)
    data = np.asarray(data)
    n_lines = len(y)

    assert data.ndim == 2, ValueError("data must be 2-dimensional (%d-d data given)" % data.ndim)
    assert data.shape == (len(y),len(x)), \
        ValueError("data (%d,%d) must have dimensions (len(y),len(x)) = (%d,%d)" \
                    % (*data.shape,len(y),len(x)))

    # If y is numeric, use it to plot y-axis; otherwise (eg if string labels) use 0:n_lines-1
    y_plot = y if isnumeric(y) else np.arange(n_lines)

    if ax is None: ax = plt.gca()

    # Sort any keyword args to their appropriate plotting object
    axes_args, plot_args = _hash_kwargs(kwargs, [AXES_PARAMS, PLOT_PARAMS])
    # Merge any input parameters with default values
    xlim = (x.min(),x.max())
    ylim = (y_plot[0]-1,y_plot[-1]+1)
    axes_args = _merge_dicts(dict(xlim=xlim, ylim=ylim), axes_args)
    plot_args = _merge_dicts(dict(linewidth=1), plot_args)
    # Set plotting colors, including defaults
    color = _set_plot_colors(color, n_lines)

    # Scale data so max range of data = <scale>*offset btwn lines on plot
    max_val = np.abs(data).max()
    data = scale * data / max_val

    ax.set(**axes_args) # Set axes parameters

    # Plot event markers (if input)
    if events is not None:
        if callable(events):    events()
        else:                   plot_markers(events, axis='x', ax=ax,
                                             xlim=axes_args['xlim'], ylim=axes_args['ylim'])

    # Plot each line plot (eg channel) in data with appropriate offset
    lines = []
    for j in range(n_lines):
        offset = y_plot[n_lines - (j+1)] if origin == 'upper' else y_plot[j]
        # offset = n_lines - (j+1) if origin == 'upper' else j
        tmp_lines = ax.plot(x, data[j,:] + offset, color=color[j], **plot_args)
        lines.append(tmp_lines)

    ax.set_yticks(y_plot)
    ax.set_yticklabels(y if origin == 'lower' else np.flip(y))

    return lines, ax


# =============================================================================
# Plotting utilities
# =============================================================================
def full_figure(**kwargs):
    """
    Create full-screen figure

    Wrapper around :func:`plt.figure` that sets size to full screen.

    Parameters
    ----------
    **kwargs
        Any keyword args passed directly to :func:`plt.figure`

    Returns
    -------
    fig : Figure object
        Output of :func:`plt.figure`
    """
    if 'frameon' not in kwargs: kwargs.update(frameon=True) # WAS False (changed due to MPL 3.1.0)
    fig = plt.figure(**kwargs)
    _maximize_figure()
    return fig


def savefig(filename, fig=None, figsize=(11.0,8.5), dpi=500, makedir=True, **kwargs):
    """
    Save figure to file in (more-or-less) WYSIWYG manner, generate target directory if missing
    
    Wrapper around fig.savefig()

    Parameters
    ----------
    filename : str
        Full-path filename to save figure into. If no file extension included, by default
        we add .png (to save a PNG file).

    fig : Pyplot Figure object, default: plt.gcf()
        Figure to save. Defaults to current figure.

    figsize : 2-tuple of float, default: default: (11.0,8.5)
        Figure dimension (width, height) in inches. Defaults to standard 8.5 x 11 in portrait.

    dpi : float, default: 500
        Resolution of saved figure in dots per inch.

    makedir : bool, default, True
        If True, creates requested directory to save figure into if missing

    References
    ----------
    https://stackoverflow.com/questions/45515320/matplotlib-savefig-fullscreen
    """
    if fig is None: fig = plt.gcf()

    # Create full path to save figure file into if doesn't already exist
    path, file = os.path.split(filename)
    if not os.path.exists(path):
        if makedir:
            os.makedirs(path)
        else:
            raise RuntimeError("Directory '%s does not exist. "
                               "Please make it or set `makedir`=True." % path)

    # If filename has no extension, make it PNG by defaults
    _, ext = os.path.splitext(file)
    if ext == '':   filename = filename + '.png'

    # Set size of figure to save and save it
    fig.set_size_inches(figsize, forward=False)
    fig.savefig(filename, dpi=dpi, **kwargs)


def make_colormap(name=None, colors=None, register=None, **kwargs):
    """
    Create a custom colormap and register its name for later convenient use
    
    Parameters
    ----------
    name : str
        Name of colormap to create. If `register` is True, name will be registered as a
        matplotlib colormap, so later you can invoke it using cmap=`name`.        
        
    colors : callable or dict or array-like
        Specifies the colors in custom colormap in one of three ways:
        
        (1) callable : `colors` is input as a function/lambda that generates
            `colors` under one of the two following specifications...
            
        (2) dict : Keys = 'red', 'green', 'blue', and (optionally) 'alpha'.
            Values for each = (n_segments-1,3) array-like of floats in range (0,1). These are
            anchor points for each color, and segments of colormap will be linearly interpolated
            between each to generate a full colormap. The first of the 3 values in each row
            determines where the anchor point lies in the colormap (0-1 ~ lowest to highest point).
            Color segements are interpolated from the 3rd value in one row to the 2nd value in 
            the subsequent row (see LinearSegmentedColormap ref below for full explanation).
            Colormap generated using :func:`matplotlib.colors.LinearSegmentedColormap`.

        (3) array-like : List of Matplotlib color specifications, or an equivalent 
            (n_colors,3=RGB) or (n_colors,4=RGBA) array. Specifies each color in entire 
            colormap. Colormap generated using :func:`matplotlib.colors.ListedColormap`.

    register : bool, default: True if value given for `name`
            If True, `name` is registered as matplotlib colormap, which can later be invoked
            using cmap=`name`.
            
    **kwargs :
        Any other keyword args passed directly to `LinearSegmentedColormap` or `ListedColormap`.
        
    Returns
    -------
    cmap : matplotlib.colors.Colormap object (LinearSegmentedColormap or ListedColormap) 
        Generated colormap        
    
    References
    ----------
    https://matplotlib.org/stable/tutorials/colors/colormap-manipulation.html
    https://matplotlib.org/stable/api/_as_gen/matplotlib.colors.LinearSegmentedColormap.html
    https://matplotlib.org/stable/api/_as_gen/matplotlib.colors.ListedColormap.html
    """
    if register is None: register = name is not None
    
    # If `colors` input as callable, run it to generate actual colors for colormap
    colors = colors() if callable(colors) else colors
    
    # If `colors` is dict, we assume it contains points to linearly interp colormap segments btwn
    if isinstance(colors,dict):        
        cmap = LinearSegmentedColormap(name=name, segmentdata=colors, **kwargs)
        
    # If `colors` is array/list, we assume it directly represents all colors in colormap    
    elif isarraylike(colors):
        cmap = ListedColormap(colors=colors, name=name, **kwargs)
    
    else:
        raise TypeError("Unsupported type <%s> set for `colors`" % type(colors))
    
    # Register colormap name for later use        
    if register: plt.register_cmap(cmap=cmap)
    
    return cmap


def colorbar(mappable=None, ax=None, size=0.05, pad=0.05, **kwargs):
    """
    Create a colorbar for given axis without messing up parent axis size (as plt.colorbar() does)

    Wrapper around `fig.colorbar`
    
    Parameters
    ----------
    mappable : matplotlib.cm.ScalarMappable object, default: ax._gci() (current artist)
        The thing this colorbar is supposed to describe (eg the output of plt.imshow()).
        Defaults to current colorable artist, if available.

    ax :  Pyplot Axis object, default: plt.gca() (current axis)
         Parent axis to plot colorbar for (eg axis with image/heatmap plot)

    size : float, default: 0.05
        Colorbar width, expressed as proportion of parent axis width

    pad : float, default: 0.05
        Distance of colorbar from parent axis, expressed as proportion of parent axis width

    **kwargs :
        Any other keyword args passed directly to `fig.colorbar`.

    Returns
    -------
    cbar : matplotlib.colorbar.Colorbar object
        Output of plt.colorbar(). Allows customization of colorbar properties.

    References
    ----------
    https://stackoverflow.com/questions/32462881/add-colorbar-to-existing-axis
    """
    # Default: Get the current colorable artist (see pyplot for details)
    if ax is None: ax = plt.gca()
    fig = ax.figure
    if mappable is None: mappable = ax._gci()

    assert mappable is not None, \
        RuntimeError('No mappable was found to use for colorbar creation. '
                     'First define a mappable such as an image (eg with imshow)')

    # Get size/position of parent axis: (x0, y0, width, height)
    bounds = ax.get_position().bounds

    # Make colorbar axis size/position based on base axis
    # x position of left side of colorbar = `pad`*(parent axis width) from right side
    xpos = bounds[0] + bounds[2] + bounds[2]*pad
    # Width of colorbar = `size`*(parent axis width)
    width   = bounds[2]*size

    # Add new axis of given position,size for colorbar
    cax = fig.add_axes([xpos, bounds[1], width, bounds[3]])
    # Create colorbar
    return fig.colorbar(mappable, cax=cax, **kwargs)


def plot_markers(values, axis='x', ax=None, xlim=None, ylim=None,
                 linecolor=[0.50,0.50,0.50], linewidth=0.5,
                 fillcolor=[0.50,0.50,0.50], fillalpha=0.2):
    """
    Plot set of markers on given axis/axes (eg to mark trial event times)

    Single point events should be input as scalars in `values`, and are plotted as
    a single line of given color and width.

    Events extending over a range/duration should be input as 2-length (start,end) tuples,
    in `values`, and are plotted as filled rectangles of given color and alpha (transparency).

    Events that reflect a central value (eg mean) +/- a range or error (eg SDs) should
    be input as 3-length (center-error,center,center+error) tuples in `values`, and are
    plotted as a solid central line with dashed error lines, in the given color and width.

    All marker types extend the full length of the opposing axis.

    NOTE: If limits are explicitly input for the axis the markers are plotted on, any marker fully
    outside of the plot limits will not be plotted, to avoid surprise expansion of the plot limits.

    Parameters
    ----------
    values : array-like, shape=(n_events,), dtype=scalars and/or 2-tuples and/or 3-tuples
        List of values (eg trial event times) on given axes to plot markers for.
        Each entry in list can be a scalar (plotted as a line), a (start,end)
        2-length tuple (plotted as a filled rectangle), or a (-err,center,+err)
        3-length tuple (plotted as solid line with surrounding dashed lines).

    axis : {'x','y','both'}, default: 'x'
        Which axis to plot markers on -- x-axis, y-axis, or both axes (eg for time x time plot)

    Returns
    -------
    ax : Axis object
        Axis plotted into.

    handles : List of Line2D, Polygon, or lists of Line2D objects
        plt.plot/fill outputs for each marker plotted, in the same order as input.
        Allows access to properties of marker lines/fills.
    """
    if isinstance(values,float) or isinstance(values,int): values = [values]
    xlim_input = xlim is not None
    ylim_input = ylim is not None
    if ax is None: ax = plt.gca()
    if xlim is None: xlim = ax.get_xlim()
    if ylim is None: ylim = ax.get_ylim()

    axis = axis.lower()
    assert axis in ['x','y','both'], ValueError("axis must be 'x'|'y'|'both'")
    axes = ['x','y'] if axis == 'both' else [axis]

    # Functions to plot lines or fills for each event in list
    def plot_single_line(value, axis):
        """ Plot scalar value as single line extending the length of the opposing axis """
        if axis == 'x':
            lines = ax.plot([value]*2, ylim, '-', color=linecolor, linewidth=linewidth)
        elif axis == 'y':
            lines = ax.plot(xlim, [value]*2, '-', color=linecolor, linewidth=linewidth)

        return lines

    def plot_fill(value, axis):
        if axis == 'x':
            patches = ax.fill([value[0],value[0],value[1],value[1]],
                              [ylim[0],ylim[1],ylim[1],ylim[0]],
                              color=fillcolor, edgecolor=None, alpha=fillalpha)
        elif axis == 'y':
            patches = ax.fill([xlim[0],xlim[1],xlim[1],xlim[0]],
                              [value[0],value[0],value[1],value[1]],
                              color=fillcolor, edgecolor=None, alpha=fillalpha)

        return patches

    def plot_three_lines(value, axis):
        """ Plot 3-tuple as 3 lines (dash,solid,dash) extending length of the opposing axis """
        lines = [None]*3
        if axis == 'x':
            lines[0] = ax.plot([value[0]]*2, ylim, '--', color=linecolor, linewidth=linewidth)
            lines[1] = ax.plot([value[1]]*2, ylim, '-', color=linecolor, linewidth=linewidth)
            lines[2] = ax.plot([value[2]]*2, ylim, '--', color=linecolor, linewidth=linewidth)
        elif axis == 'y':
            lines[0] = ax.plot(xlim, [value[0]]*2, '--', color=linecolor, linewidth=linewidth)
            lines[1] = ax.plot(xlim, [value[1]]*2, '-', color=linecolor, linewidth=linewidth)
            lines[2] = ax.plot(xlim, [value[2]]*2, '--', color=linecolor, linewidth=linewidth)

        return lines


    # Iterate thru each marker value (eg event time) and plot line or fill marker for it
    handles = []
    for value in values:
        if isinstance(value,float) or isinstance(value,int): value = [value]
        value = np.atleast_1d(value)
        if value.shape[0] == 0: continue

        # Iterate thru axes (if > 1) to plot markers on
        for axis in axes:
            # Skip plotting any markers that are entirely out of axis limits
            # Note: Only do this if axis limits are explicitly input, to avoid user confusion
            if (axis == 'x') and xlim_input:
                if (value < xlim[0]).all() or (value > xlim[1]).all(): continue
            elif (axis == 'y') and ylim_input:
                if (value < ylim[0]).all() or (value > ylim[1]).all(): continue

            # Plot lines if only scalar value (eg single unitary event time)
            if len(value) == 1:     handle = plot_single_line(value, axis)
            # Plot fill if there are 2 values (start,end) (eg event of given range or duration)
            elif len(value) == 2:   handle = plot_fill(value, axis)
            # Plot fill if there are 2 values (start,end) (eg event of given range or duration)
            elif len(value) == 3:   handle = plot_three_lines(value, axis)
            else:
                raise ValueError("Each value in values must be scalar|2-tuple|3-tuple (not len=%d)"
                                % len(value))

            handles.append(handle)

    return ax, handles


# =============================================================================
# Helper functions
# =============================================================================
def _hash_kwargs(args_dict, attr_lists):
    """
    Given a dict of keyword args input into a plotting function, determine which
    are attributes of each of a given set of plot objects (eg Axes, Lines2D, etc.).

    Parameters
    ----------
    args_dict : dict {str:value}
        List of all keyword args (name:value pairs) input into a plotting function

    attr_lists : List of lists
        Set of lists of attributes of plotting objects that keywords args
        may be attributes of. Function matches each args to its proper attribute list
        (and thus, to its corresponding plotting object).

    Returns
    -------
    **hashed_attrs : tuple of dicts {str:value}
        Each dict contains name:value pairs for all keyword args
        corresponding to each attr_list/plotting object (ie 1st output = args
        corresponding to `attr_lists[0]`, 2nd output ~ `attr_lists[1]`, etc.)

        Any keyword args not matched to any `attr_list` will raise an error.
    """
    # Create list of empty dictionaries to hash arguments into
    hashed_attrs = [{} for _ in range(len(attr_lists))]

    # Step thru each key,value pair in arguments dictionary
    for key,value in args_dict.items():
        # Step thru each passed list of object attributes to hash arguments into,
        # determining if key can be found in its attribute list
        found = False
        for i_list,attr_list in enumerate(attr_lists):
            # If we found a match, save k,v pair into corresponding output hashed dict
            # and break out of for att_lists loop
            if key in attr_list:
                hashed_attrs[i_list].update({key:value})
                found = True
                break

        # If we failed to find a match in any attribute list, raise an error
        if found == False:
            raise AttributeError("Incorrect or misspelled variable in keyword args: %s" % key)

    return tuple(hashed_attrs)


def _set_plot_colors(color, n_plot_objects):
    """ Set plotting colors, including defaults and expanding colors to # of plot objects """
    # If no color set, set default = ['C0','C1',...,'CN'] = default matplotlib plotting color order
    if color is None:
        color = ['C'+str(j) for j in range(n_plot_objects)]

    # Otherwise, ensure number of colors matches number of plotting objects, expanding if necessary
    else:
        color = np.atleast_1d(color)
        # If a RBG triplet is input, enclose it in an outer array, to simplify downstream code
        if ((len(color) == 3) and (n_plot_objects != 3)): color = [color]

        # If only 1 color input, replicate it for all plot objects (ie lines,dots,fills,etc.)
        if (len(color) == 1) and (n_plot_objects != 1):
            color = np.tile(color, (n_plot_objects,))
        else:
            assert len(color) == n_plot_objects, \
                ValueError("Color must have one value per plot obect (line/fill/etc)" \
                           " or a single value that is used for all plot objects" \
                           " (%d colors/%d objects)" % (len(color),n_plot_objects))

    return color


def _maximize_figure():
    """
    Maximize size of current Pyplot figure to fill full screen

    References
    ----------
    https://stackoverflow.com/a/32428266
    """
    manager = plt.get_current_fig_manager()
    
    # Method depends on which Matplotlib backend you are using
    backend = get_backend()    
    if 'qt' in backend.lower():     # QT backend
        manager.resize(manager.window.maximumWidth(), manager.window.maximumHeight())
    elif 'tk' in backend.lower():   # TkAgg backend
        manager.resize(*manager.window.maxsize())
    elif 'wx' in backend.lower():   # WX backend
        manager.frame.Maximize(True)
    else:
        warn("Unsupported Matplotlib backend '%s'. Could not maximize figure." % backend)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Functions for loading from and saving to Matlab MAT files

Overview
--------
Seamlessly loads all Matlab .mat files, both older versions (which require scipy.io loaders)
and the newer HDF5-based version (which requires h5py loaders). Dispatch to appropriate loader
takes place under-the-hood, without any user input.

Matlab variables are loaded into the closest Python approximation that is also useful for data
analysis, with some user-specified options. These are summarized below (see `loadmat` for details):

=================== =============================================
MATLAB              PYTHON
=================== =============================================
array               Numpy ndarray of appropriate dtype
cell array          Numpy ndarray of `object` dtype
struct (table-like) dict or Pandas DataFrame (user's option)
struct (general)    dict
=================== =============================================

All "container" variables (structs, cell arrays) are loaded recursively, so each element
(and sub-element, etc.) is processed appropriately.

Module also contains functionality for introspection into the contents of mat files without
having to load them (`whomat`), and for saving variables back out to mat files with proper
conversion to types encodable in mat files (`savemat`).


Function list
-------------
- loadmat : Loads variables from any Matlab MAT file. Also aliased as 'load'.
- savemat : Saves given variables into a MAT file. Also aliased as 'save'.
- whomat :  Lists all variables in any MAT file. Also aliased as 'who'.

Dependencies
------------
- h5py :    Python interface to the HDF5 binary data format (used for v7.3 MAT files)
- hdf5storage : Utilities to write Python types to HDF5 files, including v7.3 MAT files.

Function reference
------------------
"""
# Created on Mon Mar 12 17:20:26 2018
#
# @author: sbrincat
import sys
from copy import deepcopy
import numpy as np
import pandas as pd

from spynal.matIO.matIO_7 import _load7, _who7, _save7
from spynal.matIO.matIO_73 import _load73, _who73, _save73
from spynal.matIO.helpers import _parse_typemap, _parse_extract_items, _get_matfile_version, \
                                 _variables_to_mat


# =============================================================================
# Matfile loading/introspection functions
# =============================================================================
def loadmat(filename, variables=None, typemap=None, asdict=False, extract_items=None,
            order='Matlab', verbose=True):
    """
    Load variables from a given MAT file and return them in appropriate Python types

    Handles both older (v4-v7) and newer (v7.3) versions of MAT files,
    transparently to the user.

    Variables returned individually or in a dict, where each variable maps to key/value pair

    Returned variable types are logical Python equivalents of Matlab types:

    =================== ===============================================
    MATLAB              PYTHON
    =================== ===============================================
    double/single       float
    int                 int
    char,string         str
    logical             bool
    array               Numpy ndarray of appropriate dtype
    cell array          Numpy ndarray of `object` dtype
    struct (table-like) dict or Pandas DataFrame (depends on `typemap`)
    struct (general)    dict
    =================== ===============================================

    Single-element Matlab arrays are converted to the contained item type (eg float/int/str) by
    default (behavior can be changed using `extract_items` argument)

    NOTE: Some proprietary or custom Matlab variables cannot be loaded, including:
    table/timetable, datetime, categorical, function_handle, map container, any custom object class

    Parameters
    ----------
    filename : str
        Full-path name of MAT file to load from

    variables : list of str, default: <all variables in file>
        Names of all variables to load

    typemap : dict {str:str}, default: {'array':'array', 'cell':'array', 'struct':'dict'}
        Maps names of Matlab variables or variable types to returned Python variable types.
        Currently alternative options only supported for table-like structs, which can
        return either as 'dict' or 'dataframe' (Pandas DataFrame).

    asdict : bool, default: False
        If True, returns variables in a {'variable_name':value} dict.
        If False, returns variables separately (as tuple).

    extract_items : bool or dict, {str:bool}, default: {'array':True, 'cell':False, 'struct':True}
        All Matlab variables -- even scalar-valued ones -- are loaded as arrays. This argument
        determines whether scalar-valued variables are returned as length-1 arrays (False) or
        the single item is extracted from the array and returned as its specific dtype (True).
        
        Can be given as a single bool value to be used for *all* loaded variables or as a
        dict that maps names of Matlab variable types or specific Matlab variable names to bools.
        
        Default: extract scalar items, except for loaded Matlab cell arrays / Python object arrays,
        where it can break downstream code to have some array elements extracted, but others 
        remaining contained in (sub)arrays (eg spike timestamp lists with a single spike for some
        trials/units).

    order : str, default: 'Matlab'
        Dimension order of returned arrays. Determines how values are arranged when reshaped.
        Options:
        - 'Matlab'/'F'  : Use Matlab/Fortran dimensional ordering (column-major-compatible)
        - 'Python'/'C'  : Use Python/C dimensional ordering (row-major compatible)

    verbose : bool, default: True
        If True, prints names and shapes of all loaded variables to stdout

    Returns
    -------
    data_dict : dict {str:<variable>}
        Dictionary holding all loaded variables, mapping variable name to its value

    -or-

    vbl1,vbl2,... :
        Variables returned individually, as in 2nd example above

    Examples
    --------
    data_dict = loadmat(filename, variable, asdict=True)

    variable1, variable2, ... = loadmat(filename, variable, asdict=False)
    """
    # If variables input as string, convert to list
    if isinstance(variables,str):  variables = [variables]

    # Combine any input values for typemap with defaults
    typemap = _parse_typemap(typemap)
    extract_items = _parse_extract_items(extract_items)

    assert order.upper() in ['MATLAB','F','COL','COLMAJOR','PYTHON','C','ROW','ROWMAJOR'], \
        "<order> must be 'Matlab' or 'Python' (%s given)" % order

    version = _get_matfile_version(filename)

    # Use h5py to load v7.3 MAT-files (which are a type of hdf5 file)
    if version == 7.3:
        data = _load73(filename, variables=variables, typemap=typemap,
                       extract_items=extract_items, order=order)

    # Use scipy.io.loadmat() to load v7 and older MAT-files
    else:
        data = _load7(filename, variables=variables, typemap=typemap,
                      extract_items=extract_items, order=order)

    if variables is None: variables = list(data.keys())

    if verbose:
        for vbl in variables:
            if isinstance(data[vbl], np.ndarray):       # Numpy array variables
                vblstr = vbl + ' : numpy.array(' + \
                        ''.join('%3d,' % x for x in data[vbl].shape) + ')' + \
                        (' of type %s' % data[vbl].dtype)
            elif isinstance(data[vbl], pd.DataFrame):   # Pandas DataFrame variables
                vblstr = vbl + ' : pandas.DataFrame with columns[' + \
                        ''.join('%s,' % x for x in data[vbl].keys()) + ']'
            elif isinstance(data[vbl], dict):           # dict variables (Matlab structs)
                vblstr = vbl + ' : dict with keys[' + \
                        ''.join('%s,' % x for x in data[vbl].keys()) + ']'
            else:                                       # Scalar variables (float/int/string)
                vblstr = vbl
            vblstr = vblstr + '\n'
            print(vblstr)

    if asdict:                  return data
    elif len(variables) == 1:   return data[variables[0]]
    else:                       return tuple(data[vbl] for vbl in variables)

load = loadmat
""" Alias of :func:`loadmat`. See there for details. """


def whomat(filename, verbose=True):
    """
    Return list of variables in a given MAT file and/or print them to stdout

    Parameters
    ----------
    filename : str
        Full-path name of MAT-file to examine

    verbose : bool, default: True
        If True, prints names of all file variables to stdout

    Returns
    -------
    variables : list of str
        Names of variables in file
    """
    version = _get_matfile_version(filename)

    # Use h5py to load v7.3 MAT-files (which are a type of hdf5 file)
    if version == 7.3:  variables = _who73(filename)

    # Use scipy.io.loadmat() to load v7 and older MAT-files
    else:               variables = _who7(filename)

    if verbose: print(variables)

    return variables

who = whomat
""" Alias of :func:`whomat`. See there for details. """


# =============================================================================
# Matfile saving functions
# =============================================================================
def savemat(filename, variables, version=None, **kwargs):
    """
    Save data variables to a Matlab MAT file

    Parameters
    ----------
    filename : str
        Full-path name of MAT file to save to

    variables : dict {str:<variable>}
        Names and values of variables to save

    version : float, default: (7.3 if any variable is > 2 GB; 7 otherwise)
        Version of MAT file: 7 (older) | 7.3 (newer/HDF5).

    **kwargs
        All other keyword args passed to scipy.io.savemat()
    """
    assert (version is None) or (version in [7,7.3]), ValueError("version must be 7 or 7.3")

    # Do any necessary conversions to get all variables into matfile-compatible format
    # Note: Use deepcopy to create copy of all variables to avoid changing in caller
    variables = _variables_to_mat(deepcopy(variables))

    # If version is not set or set=7, check to make sure no variables are > 2 GB
    if version != 7.3:
        # Compute max size in memory of any variable (in GB)
        sizes = [sys.getsizeof(variables[vbl]) for vbl in variables.keys()]
        max_size = np.max(np.asarray(sizes))/(1024.0**3)
        # If any veriable is > 2GB, must use v7.3, otherwise default to v7
        if max_size >= 2:
            if version == 7: print('WARNING: Variable > 2 GB, switching to MAT file v7.3')
            version = 7.3
        else:
            version = 7

    # Use hdf5storage to save v7.3 MAT-files (which are a type of hdf5 file)
    if version == 7.3:
        _save73(filename, variables, **kwargs)

    # Use scipy.io.savemat() to save v7 MAT-files
    else:
        _save7(filename, variables, **kwargs)

save = savemat
""" Alias of :func:`savemat`. See there for details. """


# Setup so module can be run from command line using:  python matIO.py <arguments>
if __name__ == "__main__":
    loadmat(sys.argv[1])



# -*- coding: utf-8 -*-
"""
Nonparametric randomization, permutation (shuffle), and bootstrap statistics

Overview
--------
Functionality for hypothesis significance testing and confidence interval computation based on
random resampling of the observed data. This uses the data itself to generate an expected
distribution for the null hypothesis, and does not rely on any specific assumptions about
the form of the data distribution(s).

Users interested in *non-randomization* (eg rank-based) nonparametric methods are encouraged
to look at the `statsmodels <https://www.statsmodels.org/>`_ package. Users interested in traditional
parametric methods are encouraged to look at the `scipy.stats
<https://docs.scipy.org/doc/scipy/reference/stats.html>`_ submodule.

Includes tests/confints for several common data schemes:

- one-sample (are data values different from 0 or baseline?)
- paired-sample difference (are paired data observations different?)
- paired-sample association (are paired data observations correlated?)
- two-sample difference (are two groups/conditions of data different?)
- one-way difference (is there some difference between multiple groups/conditions of data?)
- two-way analysis (for data varying along 2 dims, are there diff's along each and/or interaction?)

Most significance tests include options for either permutation (shuffle) or bootstrap methods.

All functions can compute tests/confints based on random resampling of a default statistic typical
for given data scheme (eg t-statistic, F-statistic) or on any custom user-input statistic.

For data not conforming to the above schemes, there is also direct access to low-level functions
for generating samples for permutations or bootstraps in your own code.

Most functions perform operations in a mass-univariate manner. This means that
rather than embedding function calls in for loops over channels, timepoints, etc., like this::

    for channel in channels:
        for timepoint in timepoints:
            results[timepoint,channel] = compute_something(data[timepoint,channel])

You can instead execute a single call on ALL the data, labeling the relevant axis
for the computation (usually trials/observations here), and it will run in parallel (vectorized)
across all channels, timepoints, etc. in the data, like this:

``results = compute_something(data, axis)``


Function list
-------------
Hypothesis tests
^^^^^^^^^^^^^^^^
- one_sample_test :             Random-sign/bootstrap 1-sample tests (~ 1-sample t-test)

- paired_sample_test :          Permutation/bstrap paired-sample difference tests (~ paired t-test)
- paired_sample_test_labels :   Same, but with (data,labels) arg format instead of (data1,data2)
- paired_sample_association_test : Perm/bstrap paired-sample association tests (~ correlation)
- paired_sample_association_test_labels : Same, but with (data,labels) arg format

- two_sample_test :             Permutation/bootstrap for all 2-sample tests (~ 2-sample t-test)
- two_sample_test_labels :      Same, but with (data,labels) arg format instead of (data1,data2)

- one_way_test :                Permutation 1-way multi-level test (~ 1-way ANOVA/F-test)
- two_way_test :                Perm 2-way multi-level/multi-factor test (~ 2-way ANOVA/F-test)

Confidence intervals
^^^^^^^^^^^^^^^^^^^^
- one_sample_confints :         Bootstrap confidence intervals for any one-sample stat
- paired_sample_confints :      Bootstrap confidence intervals for any paired-sample stat
- two_sample_confints :         Bootstrap confidence intervals for any two-sample stat

Function reference
------------------
"""
# Created on Tue Jul 30 16:28:12 2019
#
# @author: sbrincat

# TODO  Parallelize resampling loops! (using joblib?)

from math import sqrt
import numpy as np

from spynal.utils import set_random_seed, axis_index_slices, data_labels_to_data_groups, \
                         correlation
from spynal.randstats.sampling import bootstraps
from spynal.randstats.permutation import one_sample_randomization_test, \
                                         paired_sample_permutation_test, \
                                         paired_sample_association_permutation_test, \
                                         two_sample_permutation_test, \
                                         one_way_permutation_test, two_way_permutation_test
from spynal.randstats.bootstrap import one_sample_bootstrap_test, paired_sample_bootstrap_test, \
                                       paired_sample_association_bootstrap_test, \
                                       two_sample_bootstrap_test
from spynal.randstats.utils import confint_to_indexes
from spynal.randstats.helpers import _tail_to_compare, _two_sample_data_checks, \
                                     _str_to_one_sample_stat, _str_to_two_sample_stat


# =============================================================================
# One-sample randomization tests
# =============================================================================
def one_sample_test(data, axis=0, method='randomization', mu=0, stat='t', tail='both',
                    n_resamples=10000, seed=None, return_stats=False, keepdims=True, **kwargs):
    """
    Mass univariate 1-sample test of whether any arbitrary 1-sample stat (eg mean)
    is different from a given value `mu`, often 0 (analogous to 1-sample t-test).

    Parameters
    ----------
    data : ndarray, shape=(...,n,...)
        Data to run test on

    axis : int, default: 0 (1st axis)
        Axis of data corresponding to distinct trials/observations.

    method : str
        Resampling paradigm to use for test:

        - 'randomization' : Randomization sign test in :func:`.one_sample_randomization_test`
        - 'bootstrap'     : Bootstrap test in :func:`.one_sample_bootstrap_test`

    mu : float, default: 0
        Expected value of `stat` under the null hypothesis (usually 0)

    stat : str or callable, default: 't'
        Statistic to compute and resample. Can be given as a string specifier:

        - 't'     : 1-sample t-statistic
        - 'mean'  : mean across observations

        Or as a custom function to generate resampled statistic of interest.
        Should take single array argument (data) with `axis` corresponding
        to trials/observations, and return a scalar value (for each independent
        data series if multiple given).

    tail : {'both','right','left'}, default: 'both' (2-tailed test)
        Specifies tail of test to perform:

        - 'both'  : 2-tail test -- test for abs(stat_obs) > abs(stat_resmp)
        - 'right' : right-sided 1-tail test -- tests for stat_obs > stat_resmp
        - 'left'  : left-sided 1-tail test -- tests for stat_obs < stat_resmp

    n_resamples : int, default: 10000
        Number of random resamplings to perform for test (should usually be >= 10000 if feasible)

    seed : int, default: None
        Random generator seed for repeatable results. Set=None for unseeded random numbers.

    return_stats : bool, default: False
        If True, returns p values, observed stats, and resampled stats.
        If False, only returns p values.

    keepdims : bool, default: True
        If True, retains reduced observations `axis` as length-one axes in output.
        If False, removes reduced observations `axis` from output.

    **kwargs
        All other kwargs passed directly to callable `stat` function

    Returns
    -------
    p : float or ndarray, shape=(...,[1,]...)
        p values from test. For 1d data, returned as scalar value.
        For n-d data, it has same shape as data, with `axis` reduced to length 1
        if `keepdims` is True, or with `axis` removed  if `keepdims` is False.

    stat_obs : float or ndarray, shape=(...,[1,]...), optional
        Statistic values for actual observed data. Same shape as `p`.

    stat_resmp : ndarray, shape=(...,n_resamples-1,...), optional
        Distribution of statistic values for all resamplings of data.
        Same size as data, but `axis` has length n_resamples-1.

    Examples
    --------
    p = one_sample_test(data, return_stats=False)

    p, stat_obs, stat_resmp = one_sample_test(data, return_stats=True)
    """
    method = method.lower()

    if method in ['randomization','permutation','sign']:
        test_func = one_sample_randomization_test
    elif method == 'bootstrap':
        test_func = one_sample_bootstrap_test
    else:
        raise ValueError("Unsupported test type '%s'. Use 'randomization' or 'bootstrap'" % method)

    return test_func(data, axis=axis, mu=mu, stat=stat, tail=tail,
                     n_resamples=n_resamples, seed=seed, return_stats=return_stats,
                     keepdims=keepdims, **kwargs)


# =============================================================================
# Paired-sample difference randomization tests
# =============================================================================
def paired_sample_test(data1, data2, axis=0, method='permutation', d=0, stat='t', tail='both',
                       n_resamples=10000, seed=None, return_stats=False, keepdims=True, **kwargs):
    """
    Mass univariate paired-sample test of whether any arbitrary statistic (eg mean)
    differs between paired samples (analogous to paired-sample t-test)

    Parameters
    ----------
    data1,data2 : ndarray, shape=(...,n,...)
        Data from two groups to compare. Shape is arbitrary, but must be same for data1,2.

    axis : int, default: 0 (1st axis)
        Axis of data corresponding to distinct trials/observations.

    method : str
        Resampling paradigm to use for test:

        - 'permutation' : Permutation test in :func:`.paired_sample_permutation_test`
        - 'bootstrap' : Bootstrap test in :func:`.paired_sample_bootstrap_test`

    d : float, Default: 0
        Expected value of `stat` under null distribution (usually 0)

    stat : str or callable, default: 't'
        Statistic to compute and resample. Can be given as a string specifier:

        - 't' : paired t-statistic
        - 'mean'/'meandiff' : mean of pair differences

        Or as a custom function to generate resampled statistic of interest.
        Should take single array argument (equal to differences between
        paired samples, with `axis` corresponding to trials/observations)
        and return a scalar value for each independent data series.
        NOTE: Custom function should take *single array* = data1 - data2

    tail : {'both','right','left'}, default: 'both' (2-tailed test)
        Specifies tail of test to perform:

        - 'both'  : 2-tail test -- test for abs(stat_obs) > abs(stat_resmp)
        - 'right' : right-sided 1-tail test -- tests for stat_obs > stat_resmp
        - 'left'  : left-sided 1-tail test -- tests for stat_obs < stat_resmp

    n_resamples : int, default: 10000
        Number of random resamplings to perform for test (should usually be >= 10000 if feasible)

    seed : int, default: None
        Random generator seed for repeatable results. Set=None for unseeded random numbers.

    return_stats : bool, default: False
        If True, returns p values, observed stats, and resampled stats.
        If False, only returns p values.

    keepdims : bool, default: True
        If True, retains reduced observations `axis` as length-one axes in output.
        If False, removes reduced observations `axis` from output.

    **kwargs
        All other kwargs passed directly to callable `stat` function

    Returns
    -------
    p : float or ndarray, shape=(...,[1,]...)
        p values from test. For 1d data, returned as scalar value.
        For n-d data, it has same shape as data, with `axis` reduced to length 1
        if `keepdims` is True, or with `axis` removed  if `keepdims` is False.

    stat_obs : float or ndarray, shape=(...,[1,]...), optional
        Statistic values for actual observed data. Same shape as `p`.

    stat_resmp : ndarray, shape=(...,n_resamples-1,...), optional
        Distribution of statistic values for all resamplings of data.
        Same size as data, but `axis` has length n_resamples-1.
    """
    method = method.lower()

    if method in ['randomization','permutation','sign']:
        test_func = paired_sample_permutation_test
    elif method == 'bootstrap':
        test_func = paired_sample_bootstrap_test
    else:
        raise ValueError("Unsupported test type '%s'. Should be 'permutation' or 'bootstrap'"
                         % method)

    return test_func(data1, data2, axis=axis, d=d, stat=stat, tail=tail,
                     n_resamples=n_resamples, seed=seed, return_stats=return_stats,
                     keepdims=keepdims, **kwargs)


def paired_sample_test_labels(data, labels, axis=0, method='permutation', groups=None, **kwargs):
    """
    Alternative interface to paired_sample_test() that allows arguments of form (data,labels)
    instead of (data1,data2)

    Only parameters differing from :func:`paired_sample_test` are described here.

    Parameters
    ----------
    data : ndarray, shape=(...,N,...).
        Data from *both* groups to run test on.
        Arbitrary shape, but both groups must have the same n (n1 = n2 = N/2).

    labels : array-like, shape=(N,)
        Group labels for each observation (trial), identifying which group/condition
        each observation belongs to.

    groups : array-like, shape=(n_groups,), optional, default: np.unique(labels)
        List of labels for each group (condition). Used to test only a subset (pair)
        of multi-value labels.
    """
    data1, data2 = data_labels_to_data_groups(data, labels, axis=axis, groups=groups, max_groups=2)
    return paired_sample_test(data1, data2, axis=axis, method=method, **kwargs)


# =============================================================================
# Paired-sample association (correlation) randomization tests
# =============================================================================
def paired_sample_association_test(data1, data2, axis=0, method='permutation', stat='r',
                                   tail='both', n_resamples=10000, seed=None, return_stats=False,
                                   keepdims=True, **kwargs):
    """
    Mass bivariate test of association (eg correlation) between two paired samples

    Parameters
    ----------
    data1,data2 : ndarray, shape=(...,n,...)
        Data from two groups to compare. Shape is arbitrary, but must be same for data1,2.

    axis : int, default: 0 (1st axis)
        Axis of data corresponding to distinct trials/observations.

    method : str, default: 'permutation'
        Resampling paradigm to use for test:

        - 'permutation'   : Permutation test in :func:`.paired_sample_association_permutation_test`
        - 'bootstrap'     : Bootstrap test in :func:`.paired_sample_association_bootstrap_test`

    stat : str or callable, default: 'r'
        Statistic to compute and resample. Can be given as a string specifier:

        - 'r'/'pearson'   : Standard Pearson product-moment correlation
        - 'rho'/'spearman': Spearman rank correlation

        Or as a custom function to generate resampled statistic of interest.
        Should take two array arguments (data1,data2) with `axis` corresponding
        to trials/observations and return a scalar value for each independent
        data series.

    tail : {'both','right','left'}, default: 'both' (2-tailed test)
        Specifies tail of test to perform:

        - 'both'  : 2-tail test -- test for abs(stat_obs) > abs(stat_resmp)
        - 'right' : right-sided 1-tail test -- tests for stat_obs > stat_resmp
        - 'left'  : left-sided 1-tail test -- tests for stat_obs < stat_resmp

    n_resamples : int, default: 10000
        Number of random resamplings to perform for test (should usually be >= 10000 if feasible)

    seed : int, default: None
        Random generator seed for repeatable results. Set=None for unseeded random numbers.

    return_stats : bool, default: False
        If True, returns p values, observed stats, and resampled stats.
        If False, only returns p values.

    keepdims : bool, default: True
        If True, retains reduced observations `axis` as length-one axes in output.
        If False, removes reduced observations `axis` from output.

    **kwargs
        All other kwargs passed directly to callable `stat` function

    Returns
    -------
    p : float or ndarray, shape=(...,[1,]...)
        p values from test. For 1d data, returned as scalar value.
        For n-d data, it has same shape as data, with `axis` reduced to length 1
        if `keepdims` is True, or with `axis` removed  if `keepdims` is False.

    stat_obs : float or ndarray, shape=(...,[1,]...), optional
        Statistic values for actual observed data. Same shape as `p`.

    stat_resmp : ndarray, shape=(...,n_resamples-1,...), optional
        Distribution of statistic values for all resamplings of data.
        Same size as data, but `axis` has length n_resamples-1.
    """
    method = method.lower()

    if method in ['randomization','permutation','sign']:
        test_func = paired_sample_association_permutation_test
    elif method == 'bootstrap':
        test_func = paired_sample_association_bootstrap_test
    else:
        raise ValueError("Unsupported test type '%s'. Should be 'permutation' or 'bootstrap'"
                         % method)

    return test_func(data1, data2, axis=axis, stat=stat, tail=tail,
                     n_resamples=n_resamples, seed=seed, return_stats=return_stats,
                     keepdims=keepdims, **kwargs)


def paired_sample_association_test_labels(data, labels, axis=0, method='permutation', groups=None,
                                          **kwargs):
    """
    Alternative interface to paired_sample_association_test() that allows arguments of form
    (data,labels) instead of (data1,data2)

    Only parameters differing from :func:`paired_sample_association_test` are described here.

    Parameters
    ----------
    data : ndarray, shape=(...,N,...).
        Data from *both* groups to run test on.
        Arbitrary shape, but both groups must have the same n (n1 = n2 = N/2).

    labels : array-like, shape=(N,)
        Group labels for each observation (trial), identifying which group/condition
        each observation belongs to.

    groups : array-like, shape=(n_groups,), optional, default: np.unique(labels)
        List of labels for each group (condition). Used to test only a subset (pair)
        of multi-value labels.
    """
    data1, data2 = data_labels_to_data_groups(data, labels, axis=axis, groups=groups, max_groups=2)
    return paired_sample_association_test(data1, data2, axis=axis, method=method, **kwargs)


# =============================================================================
# Two-sample randomization tests
# =============================================================================
def two_sample_test(data1, data2, axis=0, method='permutation', stat='t', tail='both',
                    n_resamples=10000, seed=None, return_stats=False, keepdims=True, **kwargs):
    """
    Mass univariate two-sample test of whether any arbitrary statistic
    differs between two non-paired samples (analogous to 2-sample t-test)

    Parameters
    ----------
    data1 : ndarray, shape=(...,n1,...)
        Data from one group to compare

    data2 : ndarray, shape=(...,n2,...)
        Data from a second group to compare.
        Need not have the same n as data1, but all other dim's must be same size/shape.

    axis : int, default: 0 (1st axis)
        Axis of data corresponding to distinct trials/observations.

    method : str
        Resampling paradigm to use for test:

        - 'permutation' : Permutation test in :func:`.two_sample_permutation_test`
        - 'bootstrap' : Bootstrap test in :func:`.two_sample_bootstrap_test`

    d : float, Default: 0
        Expected value of `stat` under null distribution (usually 0)

    stat : str or callable, default: 't'
        Statistic to compute and resample. Can be given as a string specifier:

        - 't' : 2-sample t-statistic
        - 'meandiff' : group difference in means

        Or as a custom function to generate resampled statistic of interest.
        Should take single array argument (equal to differences between
        paired samples, with `axis` corresponding to trials/observations)
        and return a scalar value for each independent data series.

    tail : {'both','right','left'}, default: 'both' (2-tailed test)
        Specifies tail of test to perform:

        - 'both'  : 2-tail test -- test for abs(stat_obs) > abs(stat_resmp)
        - 'right' : right-sided 1-tail test -- tests for stat_obs > stat_resmp
        - 'left'  : left-sided 1-tail test -- tests for stat_obs < stat_resmp

    n_resamples : int, default: 10000
        Number of random resamplings to perform for test (should usually be >= 10000 if feasible)

    seed : int, default: None
        Random generator seed for repeatable results. Set=None for unseeded random numbers.

    return_stats : bool, default: False
        If True, returns p values, observed stats, and resampled stats.
        If False, only returns p values.

    keepdims : bool, default: True
        If True, retains reduced observations `axis` as length-one axes in output.
        If False, removes reduced observations `axis` from output.

    **kwargs
        All other kwargs passed directly to callable `stat` function

    Returns
    -------
    p : float or ndarray, shape=(...,[1,]...)
        p values from test. For 1d data, returned as scalar value.
        For n-d data, it has same shape as data, with `axis` reduced to length 1
        if `keepdims` is True, or with `axis` removed  if `keepdims` is False.

    stat_obs : float or ndarray, shape=(...,[1,]...), optional
        Statistic values for actual observed data. Same shape as `p`.

    stat_resmp : ndarray, shape=(...,n_resamples-1,...), optional
        Distribution of statistic values for all resamplings of data.
        Same size as data, but `axis` has length n_resamples-1.
    """
    method = method.lower()

    if method in ['randomization','permutation']:   test_func = two_sample_permutation_test
    elif method == 'bootstrap':                     test_func = two_sample_bootstrap_test
    else:
        raise ValueError("Unsupported test type '%s'. Use 'permutation' or 'bootstrap'" % method)

    return test_func(data1, data2, axis=axis, stat=stat, tail=tail,
                     n_resamples=n_resamples, seed=seed, return_stats=return_stats,
                     keepdims=keepdims, **kwargs)


def two_sample_test_labels(data, labels, axis=0, method='permutation', groups=None, **kwargs):
    """
    Altenative interface to two_sample_test() that allows arguments of form (data,labels)
    instead of (data1,data2)

    Only parameters differing from :func:`two_sample_test` are described here.

    Parameters
    ----------
    data : ndarray, shape=(...,N,...).
        Data from *both* groups to run test on.
        Arbitrary shape, but both groups must have the same n (n1 = n2 = N/2).

    labels : array-like, shape=(N,)
        Group labels for each observation (trial), identifying which group/condition
        each observation belongs to.

    groups : array-like, shape=(n_groups,), optional, default: np.unique(labels)
        List of labels for each group (condition). Used to test only a subset (pair)
        of multi-value labels.
    """
    data1, data2 = data_labels_to_data_groups(data, labels, axis=axis, groups=groups, max_groups=2)
    return two_sample_test(data1, data2, axis=axis, method=method, **kwargs)


# =============================================================================
# One-way/Two-way randomization tests
# =============================================================================
def one_way_test(data, labels, axis=0, method='permutation', stat='F', tail='right',
                 groups=None, n_resamples=10000, seed=None, return_stats=False,
                 keepdims=True, **kwargs):
    """
    Mass univariate test on any arbitrary 1-way statistic with
    multiple groups/levels (analogous to F-test in a 1-way ANOVA)

    Wrapper around functions for specific one-way tests. See those for details.

    Parameters
    ----------
    data : ndarray, shape=(...,N,...)
        Data to run test on

    labels : array-like, shape=(N,)
        Group labels for each observation (trial), identifying which group (factor level)
        each observation belongs to.

    axis : int, default: 0 (1st axis)
        Axis of data corresponding to distinct trials/observations.

    method : str
        Resampling paradigm to use for test. Currently only 'permutation' implemented.

    stat : str or callable, default: 'F'
        Statistic to compute and resample. Can be given as a string specifier:
        'F' : F-statistic

        Or as a custom function to generate resampled statistic of interest.
        Should take data array (data) with <axis> corresponding to
        trials/observations and labels arguments (labels) and return
        a scalar value for each independent data series.

    tail : {'both','right','left'}, default: 'right' (1-tailed test)
        Specifies tail of test to perform:

        - 'both'  : 2-tail test -- test for abs(stat_obs) > abs(stat_resmp)
        - 'right' : right-sided 1-tail test -- tests for stat_obs > stat_resmp
        - 'left'  : left-sided 1-tail test -- tests for stat_obs < stat_resmp

        Note: For F-test, only right-tailed test really makes sense bc F distn
        only has positive values and right-sided tailed

    groups : array-like, shape=(n_groups,), optional, default: np.unique(labels)
        List of labels for each group (condition). Used to test only a subset of labels.

    n_resamples : int, default: 10000
        Number of random resamplings to perform for test (should usually be >= 10000 if feasible)

    seed : int, default: None
        Random generator seed for repeatable results. Set=None for unseeded random numbers.

    return_stats : bool, default: False
        If True, returns p values, observed stats, and resampled stats.
        If False, only returns p values.

    keepdims : bool, default: True
        If True, retains reduced observations `axis` as length-one axes in output.
        If False, removes reduced observations `axis` from output.

    **kwargs
        All other kwargs passed directly to callable `stat` function

    Returns
    -------
    p : float or ndarray, shape=(...,[1,]...)
        p values from test. For 1d data, returned as scalar value.
        For n-d data, it has same shape as data, with `axis` reduced to length 1
        if `keepdims` is True, or with `axis` removed  if `keepdims` is False.

    stat_obs : float or ndarray, shape=(...,[1,]...), optional
        Statistic values for actual observed data. Same shape as `p`.

    stat_resmp : ndarray, shape=(...,n_resamples-1,...), optional
        Distribution of statistic values for all resamplings of data.
        Same size as data, but `axis` has length n_resamples-1.
    """
    method = method.lower()

    if method in ['randomization','permutation','sign']:    test_func = one_way_permutation_test
    else:
        raise ValueError("Only 'permutation' method currently supported")

    return test_func(data, labels, axis=axis, stat=stat, tail=tail, groups=groups,
                     n_resamples=n_resamples, seed=seed, return_stats=return_stats,
                     keepdims=keepdims, **kwargs)


def two_way_test(data, labels, axis=0, method='permutation', stat='F', tail='right', groups=None,
                             n_resamples=10000, seed=None, return_stats=False, keepdims=True, **kwargs):
    """
    Mass univariate test on any arbitrary 2-way statistic with
    multiple groups/levels (analogous to F-test in a 2-way ANOVA)

    Parameters
    ----------
    data : ndarray, shape=(...,N,...)
        Data to run test on

    labels : array-like, shape=(n,n_terms=2|3) array-like
        Group labels for each observation, identifying which group (factor level)
        each observation belongs to, for each model term. First two columns correspond
        to model main effects; optional third column corresponds to interaction term.

    axis : int, default: 0 (1st axis)
        Axis of data corresponding to distinct trials/observations.

    method : str
        Resampling paradigm to use for test. Currently only 'permutation' implemented.

    stat : str or callable, default: 'F'
        Statistic to compute and resample. Can be given as a string specifier:
        'F' : F-statistic

        Or as a custom function to generate resampled statistic of interest.
        Should take data array (data) with <axis> corresponding to
        trials/observations and labels arguments (labels) and return
        a scalar value for each independent data series.

    tail : {'both','right','left'}, default: 'right' (1-tailed test)
        Specifies tail of test to perform:

        - 'both'  : 2-tail test -- test for abs(stat_obs) > abs(stat_resmp)
        - 'right' : right-sided 1-tail test -- tests for stat_obs > stat_resmp
        - 'left'  : left-sided 1-tail test -- tests for stat_obs < stat_resmp

        Note: For F-test, only right-tailed test really makes sense bc F distn
        only has positive values and right-sided tailed

    groups : array_like, shape=(n_terms,) of [array-like, shape=(n_groups(term),)], default: all
        List of group labels to use for each for each model term.
        Used to test only a subset of labels. Default to using all values in `labels`.

    n_resamples : int, default: 10000
        Number of random resamplings to perform for test (should usually be >= 10000 if feasible)

    seed : int, default: None
        Random generator seed for repeatable results. Set=None for unseeded random numbers.

    return_stats : bool, default: False
        If True, returns p values, observed stats, and resampled stats.
        If False, only returns p values.

    keepdims : True
        NOTE: This arg not used here; only here to maintain same API with other stat func's.

    **kwargs
        All other kwargs passed directly to callable `stat` function

    Returns
    -------
    p : ndarray, shape=(...,n_terms,...)
        p values from test. Same shape as data, with `axis` reduced to length n_terms.

    stat_obs : ndarray, shape=(...,n_terms,...), optional
        Statistic values for actual observed data. Same shape as `p`.

    stat_resmp : ndarray, shape=(...,n_terms,...,n_resamples-1), optional
        Distribution of statistic values for all resamplings of data.
        Same size as data, but `axis` has length n_terms and a new axis of
        length n_resamples-1 is appended to end of array.
        NOTE: axis for resamples is different from all other functions bc
        we need to accomodate both resample and terms dimensions here.
    """
    method = method.lower()

    if method in ['randomization','permutation','sign']:    test_func = two_way_permutation_test
    else:
        raise ValueError("Only 'permutation' method currently supported")

    return test_func(data, labels, axis=axis, stat=stat, tail=tail, groups=groups,
                     n_resamples=n_resamples, seed=seed, return_stats=return_stats, **kwargs)


# =============================================================================
# Confidence intervals
# =============================================================================
def one_sample_confints(data, axis=0, stat='mean', confint=0.95, n_resamples=10000, seed=None,
                        return_stats=False, return_sorted=True, keepdims=True, **kwargs):
    """
    Mass univariate bootstrap confidence intervals of any arbitrary 1-sample stat
    (eg mean).  Analogous to SEM/parametric confidence intervals.

    Parameters
    ----------
    data : ndarray, shape=(...,n,...)
        Data to compute confints on. Arbitrary shape.

    axis : int, default: 0 (1st axis)
        Axis of data corresponding to distinct trials/observations.

    stat : str or callable, default: 'mean'
        Statistic to compute and resample. Can be given as a string specifier:
        mean'  : mean across observations

        Or as a custom function to generate resampled statistic of interest.
        Should take single array argument (data) with `axis` corresponding
        to trials/observations, and return a scalar value (for each independent
        data series if multiple given).

    confint : float, default: 0.95 (95% confidence interval)
        Confidence interval to compute, expressed as decimal value in range 0-1.
        Typical values are 0.68 (to approximate SEM), 0.95 (95% confint), and 0.99 (99%)

    n_resamples : int, default: 10000
        Number of random resamplings to perform for test (should usually be >= 10000 if feasible)

    seed : int, default: None
        Random generator seed for repeatable results. Set=None for unseeded random numbers.

    return_stats : bool, default: False
        If False, only return confidence intervals. If True, also return statistic computed
        on observed data, and full distribution of resample statistic.

    return_sorted : bool, default: True
        If True, return stat_resmp sorted by value. If False, return stat_resmp unsorted
        (ordered by resample number), which is useful if you want to keep each resampling
        for all mass-univariate data series's together.

    keepdims : bool, default: True
        If True, retains reduced observations `axis` as length-one axes in `stat_obs`.
        If False, removes reduced observations `axis` from `stat_obs`.

    **kwargs
        All other kwargs passed directly to callable `stat` function

    Returns
    -------
    confints : ndarray, shape=(...,2,...)
        Computed bootstrap confidence intervals. Same size as data, with `axis` reduced
        to length 2 = [lower,upper] confidence interval.

    stat_obs : float or ndarray, shape=(...,[1,]...), optional
        Statistic values for actual observed data. For 1d data, returned as scalar value.
        For n-d data, it has same shape as data, with `axis` reduced to length 1
        if `keepdims` is True, or with `axis` removed  if `keepdims` is False.

    stat_resmp : ndarray, shape=(...,n_resamples,...)
        Distribution of statistic values for all resamplings of data.
        Same size as data, with `axis` now having length=n_resamples.
    """
    # Wrap negative axis back into 0 to ndim-1
    if axis < 0: axis = data.ndim + axis

    # Convert string specifiers to callable functions
    stat_func    = _str_to_one_sample_stat(stat,axis)

    # Indexes into set of bootstrap resamples corresponding to given confint
    conf_indexes = confint_to_indexes(confint, n_resamples)

    ndim = data.ndim
    n = data.shape[axis]

    # Compute statistic of interest on actual observed data
    if return_stats: stat_obs = stat_func(data, **kwargs)

    # Create generators with n_resamples random bootstrap resamplings with replacement
    resamples = bootstraps(n, n_resamples, seed)

    # Compute statistic under <n_resamples> random resamplings
    stat_shape = [axlen if ax != axis else n_resamples for ax,axlen in enumerate(data.shape)]
    stat_resmp = np.empty(stat_shape)
    for i_resmp,resample in enumerate(resamples):
        # Index into <axis> of data and stat, with ':' for all other axes
        data_slices = axis_index_slices(axis, resample, ndim)
        stat_slices = axis_index_slices(axis, [i_resmp], ndim)
        # Compute statistic on resampled data
        stat_resmp[stat_slices] = stat_func(data[data_slices], **kwargs)

    # Sort boostrap resampled stats and extract confints from them
    stat_slices = axis_index_slices(axis, conf_indexes, ndim)
    if return_stats and not return_sorted:
        # Sort copy of stat_resmp, so we can return original unsorted version
        stat_resmp_sorted = np.sort(stat_resmp, axis=axis)
        # Extract lower,upper confints from resampled and sorted stats
        confints = stat_resmp_sorted[stat_slices]
    else:
        stat_resmp.sort(axis=axis)     # Sort resample stats in place
        # Extract lower,upper confints from resampled and sorted stats
        confints = stat_resmp[stat_slices]

    # For vector-valued data, extract value from scalar array -> float for output
    if return_stats and (stat_obs.size == 1): stat_obs = stat_obs.item()
    elif not keepdims: stat_obs = stat_obs.squeeze(axis=axis)

    if return_stats:    return confints, stat_obs, stat_resmp
    else:               return confints


def paired_sample_confints(data1, data2, axis=0, stat='mean', confint=0.95, n_resamples=10000,
                           seed=None, return_stats=False, return_sorted=True,
                           keepdims=True, **kwargs):
    """
    Mass univariate bootstrap confidence intervals of any arbitrary paired-sample stat
    (eg mean difference). Analogous to SEM/parametric confidence intervals.

    Parameters
    ----------
    data1,data2 : ndarray, shape=(...,n,...)
        Data from two paired groups. Shape is arbitrary, but must be same for data1,2.

    axis : int, default: 0 (1st axis)
        Axis of data corresponding to distinct trials/observations.

    stat : str or callable, default: 'mean'
        Statistic to compute and resample. Can be given as a string specifier:
        'mean'  : mean difference between paired observations

        Or as a custom function to generate resampled statistic of interest.
        Should take single array argument (equal to differences between
        paired samples, with `axis` corresponding to trials/observations)
        and return a scalar value for each independent data series.

    confint : float, default: 0.95 (95% confidence interval)
        Confidence interval to compute, expressed as decimal value in range 0-1.
        Typical values are 0.68 (to approximate SEM), 0.95 (95% confint), and 0.99 (99%)

    n_resamples : int, default: 10000
        Number of random resamplings to perform for test (should usually be >= 10000 if feasible)

    seed : int, default: None
        Random generator seed for repeatable results. Set=None for unseeded random numbers.

    return_stats : bool, default: False
        If False, only return confidence intervals. If True, also return statistic computed
        on observed data, and full distribution of resample statistic.

    return_sorted : bool, default: True
        If True, return stat_resmp sorted by value. If False, return stat_resmp unsorted
        (ordered by resample number), which is useful if you want to keep each resampling
        for all mass-univariate data series's together.

    keepdims : bool, default: True
        If True, retains reduced observations `axis` as length-one axes in `stat_obs`.
        If False, removes reduced observations `axis` from `stat_obs`.

    **kwargs
        All other kwargs passed directly to callable `stat` function

    Returns
    -------
    confints : ndarray, shape=(...,2,...)
        Computed bootstrap confidence intervals. Same size as data, with `axis` reduced
        to length 2 = [lower,upper] confidence interval.

    stat_obs : ndarray, shape=(...,1,...), optional
        Statistic values for actual observed data.
        Same size as data, with `axis` reduced to length 1.

    stat_resmp : ndarray, shape=(...,n_resamples,...)
        Distribution of statistic values for all resamplings of data.
        Same size as data, with `axis` now having length=n_resamples.
    """
    return one_sample_confints(data1 - data2, axis=axis, stat=stat, confint=confint,
                               n_resamples=n_resamples, seed=seed, return_stats=return_stats,
                               return_sorted=return_sorted, keepdims=keepdims, **kwargs)


def two_sample_confints(data1, data2, axis=0, stat='meandiff', confint=0.95, n_resamples=10000,
                        seed=None, return_stats=False, return_sorted=True,
                        keepdims=True, **kwargs):
    """
    Mass univariate bootstrap confidence intervals of any arbitrary 2-sample stat
    (eg difference in group means).  Analogous to SEM/parametric confidence intervals.

    Parameters
    ----------
    data1 : ndarray, shape=(...,n1,...)
        Data from one group to compare

    data2 : ndarray, shape=(...,n2,...)
        Data from a second group to compare.
        Need not have the same n as data1, but all other dim's must be same size/shape.

    axis : int, default: 0 (1st axis)
        Axis of data corresponding to distinct trials/observations.

    stat : str or callable, default: 'meandiff'
        Statistic to compute and resample. Can be given as a string specifier:
        'meandiff'  : difference between group means

        Or as custom function to generate resampled statistic of interest.
        Should take two array arguments (data1,2) and `axis` corresponding
        to trials/observations and return a scalar value for each independent
        data series.

    confint : float, default: 0.95 (95% confidence interval)
        Confidence interval to compute, expressed as decimal value in range 0-1.
        Typical values are 0.68 (to approximate SEM), 0.95 (95% confint), and 0.99 (99%)

    n_resamples : int, default: 10000
        Number of random resamplings to perform for test (should usually be >= 10000 if feasible)

    seed : int, default: None
        Random generator seed for repeatable results. Set=None for unseeded random numbers.

    return_stats : bool, default: False
        If False, only return confidence intervals. If True, also return statistic computed
        on observed data, and full distribution of resample statistic.

    return_sorted : bool, default: True
        If True, return stat_resmp sorted by value. If False, return stat_resmp unsorted
        (ordered by resample number), which is useful if you want to keep each resampling
        for all mass-univariate data series's together.

    keepdims : bool, default: True
        If True, retains reduced observations `axis` as length-one axes in `stat_obs`.
        If False, removes reduced observations `axis` from `stat_obs`.

    **kwargs
        All other kwargs passed directly to callable `stat` function

    Returns
    -------
    confints : ndarray, shape=(...,2,...)
        Computed bootstrap confidence intervals. Same size as data, with `axis` reduced
        to length 2 = [lower,upper] confidence interval.

    stat_obs : float orndarray, shape=(...,[1,]...), optional
        Statistic values for actual observed data. For 1d data, returned as scalar value.
        For n-d data, it has same shape as data, with `axis` reduced to length 1
        if `keepdims` is True, or with `axis` removed  if `keepdims` is False.

    stat_resmp : ndarray, shape=(...,n_resamples,...)
        Distribution of statistic values for all resamplings of data.
        Same size as data, with `axis` now having length=n_resamples.
    """
    # Wrap negative axis back into 0 to ndim-1
    if axis < 0: axis = data1.ndim + axis

    _two_sample_data_checks(data1, data2, axis)


    # Convert string specifiers to callable functions
    stat_func    = _str_to_two_sample_stat(stat,axis)

    # Indexes into set of bootstrap resamples corresponding to given confint
    conf_indexes = confint_to_indexes(confint, n_resamples)

    ndim = data1.ndim
    n1 = data1.shape[axis]
    n2 = data2.shape[axis]

    # Compute statistic of interest on actual observed data
    if return_stats: stat_obs = stat_func(data1, data2, **kwargs)

    # Create generators with n_resamples random bootstrap resamplings with replacement
    # of ints 0:n1-1 and 0:n2-1
    # Note: Seed random number generator only *once* before generating both random samples
    if seed is not None: set_random_seed(seed)
    resamples1 = bootstraps(n1,n_resamples)
    resamples2 = bootstraps(n2,n_resamples)

    # Compute statistic under <n_resamples> random resamplings
    stat_shape = [axlen if ax != axis else n_resamples for ax,axlen in enumerate(data1.shape)]
    stat_resmp = np.empty(stat_shape)
    for i_resmp,(resample1,resample2) in enumerate(zip(resamples1,resamples2)):
        # Index into <axis> of data and stat, with ':' for all other axes
        data1_slices = axis_index_slices(axis, resample1, ndim)
        data2_slices = axis_index_slices(axis, resample2, ndim)
        stat_slices = axis_index_slices(axis, [i_resmp], ndim)
        # Compute statistic on resampled data
        stat_resmp[stat_slices] = stat_func(data1[data1_slices], data2[data2_slices], **kwargs)

    # Sort boostrap resampled stats and extract confints from them
    stat_slices = axis_index_slices(axis, conf_indexes, ndim)
    if return_stats and not return_sorted:
        # Sort copy of stat_resmp, so we can return original unsorted version
        stat_resmp_sorted = np.sort(stat_resmp, axis=axis)
        # Extract lower,upper confints from resampled and sorted stats
        confints = stat_resmp_sorted[stat_slices]
    else:
        stat_resmp.sort(axis=axis)     # Sort resample stats in place
        # Extract lower,upper confints from resampled and sorted stats
        confints = stat_resmp[stat_slices]

    # For vector-valued data, extract value from scalar array -> float for output
    if return_stats and (stat_obs.size == 1): stat_obs = stat_obs.item()
    elif not keepdims: stat_obs = stat_obs.squeeze(axis=axis)

    if return_stats:    return confints, stat_obs, stat_resmp
    else:               return confints




""" Unit tests for utils.py module """
import pytest
import random
from math import pi
import numpy as np

from spynal.tests.data_fixtures import one_sample_data, two_sample_data, one_way_data, \
                                       two_way_data, MISSING_ARG_ERRS
from spynal.utils import zscore, one_sample_tstat, paired_tstat, two_sample_tstat, \
                         one_way_fstat, two_way_fstat, fano, cv, cv2, lv, \
                         set_random_seed, randperm, interp1, setup_sliding_windows, \
                         correlation, rank_correlation, \
                         gaussian, gaussian_2d, gaussian_nd, is_symmetric, is_positive_definite, \
                         index_axis, standardize_array, undo_standardize_array, \
                         data_labels_to_data_groups, data_groups_to_data_labels, \
                         iarange, unsorted_unique, isarraylike, isnumeric, isunix, ismac, ispc, \
                         object_array_equal, object_array_compare, concatenate_object_array


# =============================================================================
# Unit tests for statistics functions
# =============================================================================
def test_zscore(one_sample_data):
    """ Unit tests for zscore() function """
    data = one_sample_data
    data_orig = data.copy()
    n_trials,n_chnls = data.shape

    # Basic test of shape, value of output. Test that z-scored data has mean ~ 0, sd ~ 1
    z = zscore(data[:,0])
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert z.shape == (n_trials,)
    assert np.isclose(z.mean(), 0, rtol=1e-2, atol=1e-2)
    assert np.isclose(z.std(), 1, rtol=1e-2, atol=1e-2)

    # Test for consistent output with multi-dim array inputs and keepdims
    z = zscore(data, axis=0)
    assert z.shape == (n_trials,n_chnls)
    assert np.allclose(z.mean(axis=0), np.zeros((1,n_chnls)), rtol=1e-2, atol=1e-2)
    assert np.allclose(z.std(axis=0), np.ones((1,n_chnls)), rtol=1e-2, atol=1e-2)

    # Test for consistent output with transposed inputs
    z = zscore(data.T, axis=1)
    assert z.shape == (n_chnls,n_trials)
    assert np.allclose(z.mean(axis=1), np.zeros((n_chnls,1)), rtol=1e-2, atol=1e-2)
    assert np.allclose(z.std(axis=1), np.ones((n_chnls,1)), rtol=1e-2, atol=1e-2)

    # Test for proper function with zscore across entire array
    z = zscore(data, axis=None)
    assert z.shape == (n_trials,n_chnls)
    assert np.isclose(z.mean(), 0, rtol=1e-2, atol=1e-2)
    assert np.isclose(z.std(), 1, rtol=1e-2, atol=1e-2)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        z = zscore(data, foo=None)


@pytest.mark.parametrize('stat_func,    result',
                         [(fano,        1.95),
                          (cv,          0.43),
                          (cv2,         0.57),
                          (lv,          0.31)])
def test_snr_stats(one_sample_data, stat_func, result):
    """ Unit tests for signal-to-noise-type stats (Fano, CV, etc.) """
    data = one_sample_data
    data_orig = data.copy()
    n_datapoints, n_chnls = data.shape

    # Basic test of shape, value of output
    stat = stat_func(data[:,0], axis=0)
    print(np.round(stat,2))
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert np.isscalar(stat)
    assert np.isclose(stat, result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with multi-dim array inputs and keepdims
    stat = stat_func(data[:,:], axis=0)
    assert stat.shape == (1,n_chnls)
    assert np.isclose(stat[:,0], result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with transposed inputs
    stat = stat_func(data[:,:].T, axis=1)
    assert stat.shape == (n_chnls,1)
    assert np.isclose(stat[0,:], result, rtol=1e-2, atol=1e-2)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        stat = stat_func(data[:,0], axis=0, foo=None)


@pytest.mark.parametrize('stat_func,    result',
                         [(one_sample_tstat,    6.91),
                          (paired_tstat,        -2.97),
                          (two_sample_tstat,    -3.39),
                          (one_way_fstat,        1.79),
                          (two_way_fstat,       (3.10,24.91,0.18))])
def test_parametric_stats(two_way_data, stat_func, result):
    data, labels = two_way_data
    if stat_func is one_sample_tstat:
        data = data[:10,:]
    elif stat_func in [paired_tstat,two_sample_tstat]:
        data = data[:20,:]
        labels = labels[:20,0]
    elif stat_func is one_way_fstat:
        labels = labels[:,0]

    data_orig = data.copy()
    labels_orig = labels.copy()
    n_datapoints, n_chnls = data.shape

    extra_args = dict(axis=0)
    if stat_func is not two_way_fstat: extra_args.update(keepdims=True)

    # Basic test of shape, value of output
    if stat_func is one_sample_tstat:
        stat = stat_func(data[:,0], **extra_args)
    elif stat_func in [paired_tstat,two_sample_tstat]:
        stat = stat_func(data[labels==0,0], data[labels==1,0], **extra_args)
    else:
        stat = stat_func(data[:,0], labels, **extra_args)
    print(np.round(stat,2))
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert np.array_equal(labels,labels_orig)
    if stat_func is two_way_fstat:
        assert stat.shape == (3,)
        assert np.allclose(stat, result, rtol=1e-2, atol=1e-2)
    else:
        assert np.isscalar(stat)
        assert np.isclose(stat, result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with multi-dim array inputs and keepdims = True
    if stat_func is one_sample_tstat:
        stat = stat_func(data[:,:], **extra_args)
    elif stat_func in [paired_tstat,two_sample_tstat]:
        stat = stat_func(data[labels==0,:], data[labels==1,:], **extra_args)
    else:
        stat = stat_func(data, labels, **extra_args)
    if stat_func is two_way_fstat:
        assert stat.shape == (3,n_chnls)
        assert np.allclose(stat[:,0], result, rtol=1e-2, atol=1e-2)
    else:
        assert stat.shape == (1,n_chnls)
        assert np.isclose(stat[0,0], result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with multi-dim array inputs and keepdims = False
    if stat_func is not two_way_fstat:
        if stat_func is one_sample_tstat:
            stat = stat_func(data[:,:], axis=0, keepdims=False)
        elif stat_func in [paired_tstat,two_sample_tstat]:
            stat = stat_func(data[labels==0,:], data[labels==1,:], axis=0, keepdims=False)
        else:
            stat = stat_func(data, labels, axis=0, keepdims=False)
        assert stat.shape == (n_chnls,)
        assert np.isclose(stat[0], result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with transposed inputs
    extra_args['axis'] = 1
    if stat_func is one_sample_tstat:
        stat = stat_func(data[:,:].T, **extra_args)
    elif stat_func in [paired_tstat,two_sample_tstat]:
        stat = stat_func(data[labels==0,:].T, data[labels==1,:].T, **extra_args)
    else:
        stat = stat_func(data.T, labels, **extra_args)
    if stat_func is two_way_fstat:
        assert stat.shape == (n_chnls,3)
        assert np.allclose(stat[0,:], result, rtol=1e-2, atol=1e-2)
    else:
        assert stat.shape == (n_chnls,1)
        assert np.isclose(stat[0,0], result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with arguments swapped.
    # Stat should have same magnitude, opposite sign for 2-sample stats.
    if stat_func in [paired_tstat,two_sample_tstat]:
        stat = stat_func(data[labels==1,0], data[labels==0,0], axis=0)
        assert np.isscalar(stat)
        assert np.isclose(stat, -result, rtol=1e-2, atol=1e-2)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        if stat_func is one_sample_tstat:
            stat = stat_func(data[:,0], axis=0, foo=None)
        elif stat_func in [paired_tstat,two_sample_tstat]:
            stat = stat_func(data[labels==0,0], data[labels==1,0], axis=0, foo=None)
        else:
            stat = stat_func(data[:,0], labels, axis=0, foo=None)


@pytest.mark.parametrize('corr_type, result, result2',
                         [('pearson',   -0.33, -0.20),
                          ('spearman',  -0.36, -0.26)])
def test_correlation(two_sample_data, corr_type, result, result2):
    """ Unit tests for correlation() and rank_correlation() functions """
    data, labels = two_sample_data
    data_orig = data.copy()

    corr_func = correlation if corr_type == 'pearson' else rank_correlation

    # Basic test of shape, value of output
    r = corr_func(data[labels==0,0], data[labels==1,0], keepdims=False)
    print(np.round(r,2), type(r))
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert np.isscalar(r)
    assert np.isclose(r, result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with arguments swapped
    r = corr_func(data[labels==1,0], data[labels==0,0], keepdims=False)
    assert np.isclose(r, result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with multi-dim array inputs and keepdims
    r = corr_func(data[labels==0,:], data[labels==1,:], axis=0, keepdims=True)
    assert r.shape == (1,4)
    assert np.isclose(r[:,0], result, rtol=1e-2, atol=1e-2)

    # Test for consistent output with transposed inputs
    r = corr_func(data[labels==0,:].T, data[labels==1,:].T, axis=1, keepdims=True)
    assert r.shape == (4,1)
    assert np.isclose(r[0,:], result, rtol=1e-2, atol=1e-2)

    # Test for proper function with correlation across entire array
    r = corr_func(data[labels==0,:], data[labels==1,:], axis=None)
    assert np.isscalar(r)
    assert np.isclose(r, result2, rtol=1e-2, atol=1e-2)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        r = corr_func(data[labels==0,:], data[labels==1,:], foo=None)


# =============================================================================
# Unit tests for numerical functions
# =============================================================================
@pytest.mark.parametrize('rand_func', [np.random.rand, random.random])
def test_set_random_seed(rand_func):
    """ Unit tests for test_random_seed() """
    # Generate unseeded random number using given random number generator
    unseeded = rand_func()

    # Seed random number generators and generate seeded random number
    seed = set_random_seed()
    seeded = rand_func()

    # Generate seeded random number with same seed
    set_random_seed(seed)
    seeded2 = rand_func()

    # Check that both seeded values are identical and unseeded value is not
    assert seeded == seeded2
    assert seeded != unseeded

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        rand_func(foo=None)


def test_randperm():
    """ Unit test for randperm() function """
    n,k = 4,4
    set_random_seed(1)
    out = randperm(n,k)
    assert np.array_equal(out, [3,2,0,1])
    
    n,k = 4,2
    set_random_seed(1)
    out = randperm(n,k)
    assert np.array_equal(out, [3,2])
    
    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        randperm(n, k, foo=None)
            
def test_interp1():
    """ Unit tests for interp1() function """
    x = iarange(0,1,0.1)
    y = np.cos(2*pi*4*x)
    xinterp = iarange(0,1,0.01)
    n_interp = len(xinterp)

    y_orig = y.copy()
    result = 0.0099

    # Test basic function
    yinterp = interp1(x, y, xinterp)
    print(yinterp.shape, np.round(yinterp.mean(),4))
    assert np.array_equal(y,y_orig)     # Ensure input data isn't altered by function
    assert np.array_equal(yinterp.shape, (n_interp,))
    assert np.isclose(yinterp.mean(), result, rtol=1e-4, atol=1e-4)

    # Test function w/ 2d array data
    Y = np.tile(y[:,np.newaxis], (1,4))
    Yinterp = interp1(x, Y, xinterp, axis=0)
    assert np.array_equal(Yinterp.shape, (n_interp,4))
    assert np.allclose(Yinterp.mean(axis=0), result, rtol=1e-4, atol=1e-4)

    # Test function w/ transposed data
    print(x.shape, Y.T.shape)
    Yinterp = interp1(x, Y.T, xinterp, axis=-1)
    assert np.array_equal(Yinterp.shape, (4,n_interp))
    assert np.allclose(Yinterp.mean(axis=-1), result, rtol=1e-4, atol=1e-4)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        yinterp = interp1(x, y, xinterp, foo=None)


@pytest.mark.parametrize('func, result, result2',
                         [(gaussian,    0.9783, 2.3780),
                          (gaussian_2d, 0.8250, 1.3873),
                          (gaussian_nd, 0.0009, 0.6169)])
def test_gaussian(one_sample_data, func, result, result2):
    """ Unit tests for gaussian/gaussian_2d/gausssian_nd functions """
    data = (one_sample_data - 10)/5
    if func is gaussian:        data = data[:,0]
    elif func is gaussian_2d:   data = data[:,:2]
    elif func is gaussian_nd:   data = data[:,:3]
    data_orig = data.copy()
    n = data.shape[0]

    # Basic test of shape, value of output
    f_x = func(data)
    print(np.round(f_x[0],4))
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert f_x.shape == (n,)
    assert np.isclose(f_x[0], result, rtol=1e-4, atol=1e-4)

    # Test for consistent results with scalar input
    if func is gaussian:    f_x = func(data[0])
    else:                   f_x = func(data[0,:])
    assert np.isscalar(f_x)
    assert np.isclose(f_x, result, rtol=1e-4, atol=1e-4)

    # Test for consistent results with hand-set parameters
    params = dict(amplitude=1, baseline=0)
    if func is gaussian:
        params.update(dict(center=0, width=1))
    elif func is gaussian_2d:
        params.update(dict(center_x=0, center_y=0, width_x=1, width_y=1, orientation=0))
    elif func is gaussian_nd:
        params.update(dict(center=np.zeros((3,)), width=np.ones((3,))))
    f_x = func(data, **params)
    assert np.isclose(f_x[0], result, rtol=1e-4, atol=1e-4)

    # Test for expected results with another set of hand-set parameters
    params = dict(amplitude=2, baseline=0.5)
    if func is gaussian:
        params.update(dict(center=0.5, width=2))
    elif func is gaussian_2d:
        params.update(dict(center_x=0.5, center_y=-0.5, width_x=2, width_y=1, orientation=pi/4))
    elif func is gaussian_nd:
        params.update(dict(center=[0.5,-0.5,0.5], width=[2,1,2]))
    f_x = func(data, **params)
    print(np.round(f_x[0],4))
    assert np.isclose(f_x[0], result2, rtol=1e-4, atol=1e-4)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        f_x = func(data, foo=None)


def test_matrix_tests():
    """ Unit tests for is_symmetric, is_positive_definite """
    # Test that matrix tests pass for Identity matrix
    X = np.eye(2)
    assert is_symmetric(X)
    assert is_positive_definite(X, semi=False)

    # Test that matrix tests fail for assymmetric, non-square matrices
    X[0,1] = 2
    assert ~is_symmetric(X)
    assert ~is_positive_definite(X, semi=False)

    X = np.concatenate((np.eye(2), np.eye(2)), axis=0)
    assert ~is_symmetric(X)
    assert ~is_positive_definite(X, semi=False)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        _ = is_symmetric(X, foo=None)
    with pytest.raises(MISSING_ARG_ERRS):
        _ = is_positive_definite(X, semi=False, foo=None)


# =============================================================================
# Unit tests for data indexing and reshaping functions
# =============================================================================
def test_index_axis():
    data = np.reshape(np.arange(3,dtype=int), (3,1,1))
    idxs = np.arange(3,dtype=int)
    result = np.arange(3,dtype=int)

    print(data.shape)
    # Test that correct data is extracted for each of several axes
    for axis in range(3):
        print(axis, np.moveaxis(data,0,axis).shape)
        assert np.array_equal(index_axis(np.moveaxis(data,0,axis), axis, idxs).squeeze(), result)
        assert np.array_equal(index_axis(np.moveaxis(data,0,axis), axis-3, idxs).squeeze(), result)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        _ = index_axis(data, axis, idxs, foo=None)


@pytest.mark.parametrize('axis, target_axis', [(0,0), (1,0), (2,0), (-1,0),
                                               (0,2), (1,2), (2,2), (-1,2),
                                               (0,-1), (1,-1), (2,-1), (-1,-1)])
def test_standardize_array(axis, target_axis):
    """ Unit tests for standardize_array/undo_standardize_array """
    # Test whether standardize->unstandardize returns original array
    data = np.random.rand(3,4,5)
    data_orig = data.copy()

    data2, data_shape = standardize_array(data, axis=axis, target_axis=target_axis)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function

    data2 = undo_standardize_array(data2, data_shape, axis=axis, target_axis=target_axis)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert np.array_equal(data,data2)

    # Same test w/ 2D data array
    axis = min(axis,1)  # Reset axes from 2 -> 1
    target_axis = min(target_axis,1)

    data = np.random.rand(3,4)
    data_orig = data.copy()

    data2, data_shape = standardize_array(data, axis=axis, target_axis=target_axis)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function

    data2 = undo_standardize_array(data2, data_shape, axis=axis, target_axis=target_axis)
    assert np.array_equal(data,data_orig)     # Ensure input data isn't altered by function
    assert np.array_equal(data,data2)

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        _,_ = standardize_array(data, axis=axis, target_axis=target_axis, foo=None)
    with pytest.raises(MISSING_ARG_ERRS):
        _,_ = undo_standardize_array(data2, data_shape, axis=axis, target_axis=target_axis, foo=None)


def test_data_groups_to_data_labels(one_way_data):
    data, labels = one_way_data
    data_orig = data.copy()
    labels_orig = labels.copy()

    # Test basic function
    data1, data2, data3 = data_labels_to_data_groups(data, labels, axis=0)
    data1_orig = data1.copy()
    assert np.array_equal(data, data_orig)      # Ensure input data isn't altered by function
    assert np.array_equal(labels, labels_orig)

    data_check, labels_check = data_groups_to_data_labels(data1, data2, data3, axis=0)
    assert np.array_equal(data1, data1_orig)    # Ensure input data isn't altered by function
    assert np.array_equal(data_check, data)
    assert np.array_equal(labels_check, labels)

    # Test for consistency with transposed data
    data1, data2, data3 = data_labels_to_data_groups(data.T, labels, axis=-1)
    data_check, labels_check = data_groups_to_data_labels(data1, data2, data3, axis=-1)
    assert np.array_equal(data_check, data.T)
    assert np.array_equal(labels_check, labels)

    # Test for consistency with subgroup selection from data
    data1, data2 = data_labels_to_data_groups(data, labels, axis=0, groups=(0,1))
    data_check, labels_check = data_groups_to_data_labels(data1, data2, axis=0)
    assert np.array_equal(data_check, data[:20,:])
    assert np.array_equal(labels_check, labels[:20])

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        _,_,_ = data_labels_to_data_groups(data, labels, axis=0, foo=None)
    with pytest.raises(MISSING_ARG_ERRS):
        _,_ = data_groups_to_data_labels(data1, data2, data3, axis=0, foo=None)


# =============================================================================
# Other utility functions
# =============================================================================
@pytest.mark.parametrize('args', [(4,), (0,4), (0,4,2), (8,4,-2)])
def test_iarange(args):
    """ Unit tests for iarange() function """
    # Test that each set of arguments produces an output ending in same value
    assert iarange(*args)[-1] == 4

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        _ = iarange(*args, foo=None)


def test_unsorted_unique():
    """ Unit tests for unsorted_unique() function """
    # Test for expected output for different arguments
    x = np.asarray((3,3,2,2,1,1))
    x_orig = x.copy()
    unique = unsorted_unique(x)
    assert np.array_equal(x,x_orig)
    assert np.array_equal(unique.shape, (3,))
    assert np.array_equal(unique, (3,2,1))

    X = np.vstack((x,x))
    X_orig = X.copy()
    unique = unsorted_unique(X,axis=1)
    assert np.array_equal(X,X_orig)
    assert np.array_equal(unique.shape, (2,3))
    assert np.array_equal(unique[0,:], (3,2,1))

    unique = unsorted_unique(X.T,axis=0)
    assert np.array_equal(unique.shape, (3,2))
    assert np.array_equal(unique[:,0], (3,2,1))

    unique = unsorted_unique(X)
    assert np.array_equal(unique.shape, (3,))
    assert np.array_equal(unique, (3,2,1))

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        _ = unsorted_unique(x, foo=None)


@pytest.mark.parametrize('x, result',
                         [((1,2), True), ([1,2], True), (np.array((1,2)), True),
                          (1.2, False), (1, False), (True, False)])
def test_isarraylike(x, result):
    """ Unit tests for isarraylike() function """
    # Test that each variable type produces expected result
    assert isarraylike(x) == result

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        _ = isarraylike(x, foo=None)


@pytest.mark.parametrize('dtype, result',
                         [(np.uint16, True), (np.float64, True), (np.complex128, True),
                          (bool, False), (object, False)])
def test_isnumeric(dtype, result):
    """ Unit tests for isnumeric() function """
    # Test that each dtype produces expected result
    x = np.ones((2,), dtype=dtype)
    assert isnumeric(x) == result

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        _ = isnumeric(x, foo=None)


def test_isplatform():
    """ Unit tests for isunix(), ismac(), ispc() functions """
    # Ensure we identify 1 and only 1 platform/OS
    assert np.sum([isunix(), ismac(), ispc()]) == 1


def test_setup_sliding_windows():
    """ Unit tests for setup_sliding_windows() function """
    # Test for expected output with different argument sets
    width = 0.2
    lims = (0,1)
    wins = setup_sliding_windows(width, lims)
    assert np.array_equal(wins.shape, (5,2))
    assert np.array_equal(wins[-1,:], (0.8,1.0))

    wins = setup_sliding_windows(width, lims, step=0.1)
    assert np.array_equal(wins.shape, (9,2))
    assert np.array_equal(wins[-1,:], (0.8,1.0))

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        _ = setup_sliding_windows(width, lims, foo=None)


@pytest.mark.parametrize('axis', [0, 1])
def test_object_array_functions(axis):
    """ Unit tests for concatenate_object_array, object_array_equal, object_array_compare """
    data = [[[1, 2],    [3, 4, 5]],
            [[6, 7, 8], [9, 10]  ]]
    data = np.asarray(data, dtype=object)
    data_orig = data.copy()

    # HACK To get Numpy to make object arrays w/o being too smart and making them (2,5) arrays
    shape = (1,2) if axis == 0 else (2,1)
    result = np.empty(shape, dtype=object)
    result_true = np.empty(shape, dtype=object)
    result_false = np.empty(shape, dtype=object)
    result_true[0,0] = [True]*5
    result_false[0,0] = [False]*5
    if axis == 0:
        result[0,0] = [1,2,6,7,8]
        result[0,1] = [3,4,5,9,10]
        result_true[0,1] = [True]*5
        result_false[0,1] = [False]*5
    elif axis == 1:
        result = np.empty((2,1), dtype=object)
        result[0,0] = [1,2,3,4,5]
        result[1,0] = [6,7,8,9,10]
        result_true[1,0] = [True]*5
        result_false[1,0] = [False]*5

    data_cat = concatenate_object_array(data, axis=axis)
    print(data_cat)
    print(result)
    assert np.array_equal(data, data_orig)  # Ensure input data isn't altered by function
    assert data_cat.shape == shape

    # Test basic `object_array_equal` call (comp_func=np.array_equal, reduce_func=np.all)
    assert object_array_equal(data_cat, result)
    # Test `object_array_equal` with alternative `comp_func`
    assert object_array_equal(data_cat, result, comp_func=np.allclose)
    # Test `object_array_equal` with alternative `reduce_func`
    assert object_array_equal(data_cat, result, reduce_func=np.sum) == 2
    # Test `object_array_equal` with `reduce_func` = None
    assert np.all(object_array_equal(data_cat, result, reduce_func=None) == [True,True])

    # Test basic `object_array_compare` call (comp_func=np.equal, reduce_func=None)
    print(object_array_compare(data_cat, result), object_array_compare(data_cat, result).shape)
    print(result_true, result_true.shape)
    assert object_array_equal(object_array_compare(data_cat, result), result_true)
    # Test `object_array_compare` with alternative `comp_func`
    assert object_array_equal(object_array_compare(data_cat, result, comp_func=np.less), result_false)
    # Test `object_array_compare` with alternative `reduce_func`
    assert np.all(object_array_compare(data_cat, result, reduce_func=np.all) == [True, True])
    assert np.all(object_array_compare(data_cat, result, reduce_func=np.sum) == [5, 5])

    # Ensure that passing a nonexistent/misspelled kwarg raises an error
    with pytest.raises(MISSING_ARG_ERRS):
        _ = concatenate_object_array(data, axis=axis, foo=None)
    with pytest.raises(MISSING_ARG_ERRS):
        _ = object_array_equal(data_cat, result, foo=None)
    with pytest.raises(MISSING_ARG_ERRS):
        _ = object_array_compare(data_cat, result, foo=None)


def test_imports():
    """ Test different import methods for utils module """
    # Import entire package
    import spynal
    spynal.utils.zscore
    # Import module
    import spynal.utils as utils
    utils.zscore
    # Import specific function from module
    from spynal.utils import zscore
    zscore
